#ifndef THREADPRIMITIVES_H_
#define THREADPRIMITIVES_H_

#include "Utils.h"

#include <pthread.h>

#include <boost/function.hpp>

class Thread
{
public:
	typedef boost::function<void ()> ThreadFunc;

private:
	pthread_t m_thread;
	
	ThreadFunc m_func;

	static void* StartPThreadFunc(void* param)
	{
		Thread* thread = static_cast<Thread*>(param);

		thread->m_func();

		return 0;
	}

	public:	
	Thread(const ThreadFunc& func)
		: m_func(func)
	{
		pthread_create(&m_thread, NULL, StartPThreadFunc, this);
	};

	~Thread()
	{
		void** status = 0;

		pthread_join(m_thread, status);
	};
};

class Mutex
{
	pthread_mutex_t m_mutex;
	bool m_locked;

	friend class Condition;

	public:
	Mutex()
	: m_locked(false)
	{
		pthread_mutex_init(&m_mutex, NULL);
	};
	~Mutex()
	{
		pthread_mutex_destroy(&m_mutex);
	};

	void Lock()
	{
		pthread_mutex_lock(&m_mutex);
		ASSERT(!m_locked);
		m_locked = true;
	}
	void Unlock()
	{
		ASSERT(m_locked);
		m_locked = false;
		pthread_mutex_unlock(&m_mutex);
	}

	bool IsLocked() const
	{
		return m_locked;
	}

	class ScopedLock
	{
		Mutex& m_mutex;

		public:
		ScopedLock(Mutex& mutex)
		: m_mutex(mutex)
		{
			m_mutex.Lock();
		}
		~ScopedLock()
		{
			m_mutex.Unlock();
		}

		Mutex& GetMutex() {return m_mutex;}
	};
};

class Condition
{
	pthread_cond_t m_cond;

	public:
	Condition()
	{
		pthread_cond_init(&m_cond, NULL);
	}
	~Condition()
	{
		pthread_cond_destroy(&m_cond);
	}

	void Wait(Mutex& locked)
	{
		ASSERT(locked.m_locked);
		locked.m_locked = false;

		pthread_cond_wait(&m_cond, &locked.m_mutex);

		ASSERT(!locked.m_locked);
		locked.m_locked = true;
	}

	void Signal()
	{
		pthread_cond_broadcast(&m_cond);
	}
};

#endif
