#include "Win32GraphicsContext.h"

#include <GLee.h>

#include <string.h>

#include "R3Utils.h"

#include <GL/GL.h>

Win32GraphicsContext::Win32GraphicsContext(HWND hwnd)
: m_hwnd(hwnd)
{
	RECT clientRect;
	GetClientRect(hwnd, &clientRect);
	m_width = clientRect.right - clientRect.left;
	m_height = clientRect.bottom - clientRect.top;

	Mutex::ScopedLock initDoneLock(m_initDoneMutex);

	m_commandQueue = new CommandQueue(100, 20000, false);
	m_thread = new Thread(boost::bind<void>(&Win32GraphicsContext::Run, this));

	// Wait for the freshly created thread to finish initializing
	m_initDoneCond.Wait(m_initDoneMutex);
}

Win32GraphicsContext::~Win32GraphicsContext()
{
	delete m_commandQueue;
	delete m_thread;
}

void Win32GraphicsContext::Run()
{
	m_dc = GetDC(m_hwnd);
	
	PIXELFORMATDESCRIPTOR pfd;
	ZeroMemory(&pfd, sizeof(pfd));
	pfd.nSize = sizeof(pfd);
	pfd.nVersion = 1;
	pfd.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
	pfd.iPixelType = PFD_TYPE_RGBA;
	pfd.cColorBits = 24;
	pfd.cDepthBits = 32;
	pfd.iLayerType = PFD_MAIN_PLANE;
	int format = ChoosePixelFormat(m_dc, &pfd);
	if(format == 0)
	{
		m_commandQueue->ProcessCommands();
		ReleaseDC(m_hwnd, m_dc);
		return;
	}

	BOOL ret = SetPixelFormat(m_dc, format, &pfd);
	ASSERT(ret == TRUE);

	m_rc = wglCreateContext(m_dc);
	ASSERT(m_rc != NULL);

	ret = wglMakeCurrent(m_dc, m_rc);
	ASSERT(ret == TRUE);

	if(__GLEE_WGL_EXT_swap_control)
	{
		wglSwapIntervalEXT(1);
	}

	{
		Mutex::ScopedLock initDoneLock(m_initDoneMutex);
		m_initDoneCond.SignalAll(); // We're done initializing
	}

	m_commandQueue->ProcessCommands();

	wglMakeCurrent(NULL, NULL);
	wglDeleteContext(m_rc);
	ReleaseDC(m_hwnd, m_dc);
}

void Win32GraphicsContext::ProcessCommand(RenderFrameCommand& cmd)
{
	m_glContext.RenderGraphicsFrame(cmd.frame, cmd.options.GetProjectionMatrix());
	SwapBuffers(m_dc);
}

void Win32GraphicsContext::RenderGraphicsFrame(GraphicsFrame& frame, const GraphicsContext::RenderOptions& options)
{
	ADDCOMMAND(*m_commandQueue, RenderFrameCommand(*this, frame, options, frame.MakeFreeze()));
}

void Win32GraphicsContext::FlushWait()
{
	ASSERT(false);
}

void Win32GraphicsContext::ProcessCommand(ResizeViewportCommand& cmd)
{
	m_glContext.SetViewportSize(cmd.width, cmd.height);
}

void Win32GraphicsContext::ResizeScreen(int width, int height)
{
	m_width = width;
	m_height = height;

	ADDCOMMAND(*m_commandQueue, ResizeViewportCommand(*this, width, height));
}

int Win32GraphicsContext::GetWidth()
{
	return m_width;
}

int Win32GraphicsContext::GetHeight()
{
	return m_height;
}

