#ifndef FIFOALLOCATOR_H_
#define FIFOALLOCATOR_H_

#include <Utils.h>
#include <ThreadPrimitives.h>
#include <Exceptions.h>
#include <Allocator.h>

class FIFOAllocator : public Allocator
{
	typedef long MemoryUnit;
	MemoryUnit* m_memory;
	MemoryUnit* m_memoryEnd;

	const long m_memorySize;

	MemoryUnit* m_curStart;
	MemoryUnit* m_curEnd;

	int m_used;

	bool m_blocking;

	Mutex m_monitorLock;
	Condition m_itemRemovedCond;
	Condition m_itemAddedCond;

	public:
	FIFOAllocator(long poolSize, bool blocking=false) // When blocking is true, the thread is blocked if there is not enough memory to allocate
	: m_memorySize((poolSize+sizeof(MemoryUnit)-1)/sizeof(MemoryUnit))
	, m_blocking(blocking)
	{
		m_memory = new MemoryUnit[m_memorySize+1];
		m_memoryEnd = m_memory + m_memorySize;

		m_curStart = m_memory;
		m_curEnd = m_memory;

		m_used = 0;
	}
	~FIFOAllocator()
	{
		ASSERT(m_curStart == m_curEnd);
		delete[] m_memory;
	}

	void* Allocate(long size)
	{
		ASSERT(size > 0);
		ASSERTMSG(size < ((m_memorySize+1)*sizeof(MemoryUnit))/2, "Cannot add elements to the queue that are more than half of the size of the queue");

		Mutex::ScopedLock lock(m_monitorLock);

		int unitSize = (size+sizeof(MemoryUnit)-1)/sizeof(MemoryUnit);
		
		int actualNeededUnits = 1 + unitSize;

		// Wait until there is enough room from end to start
		{
			while((m_used > 0 && m_curEnd <= m_curStart) 
				&& m_curEnd + actualNeededUnits > m_curStart)
			{
				if(m_blocking)
				{
					m_itemRemovedCond.Wait(m_monitorLock);
				}
				else
				{
					throw OutOfMemory();
				}
			}
		}
		
		if((m_curEnd + actualNeededUnits) > m_memoryEnd)
		{
			// Restart at beginning
			*m_curEnd = -1; // Mark the end
			m_curEnd = m_memory;
		}

		// Wait until there is enough room from end to start, again (since we might have restarted to the beginning)
		{
			while((m_used > 0 && m_curEnd <= m_curStart) 
				&& m_curEnd + actualNeededUnits > m_curStart)
			{
				if(m_blocking)
				{
					m_itemRemovedCond.Wait(m_monitorLock);
				}
				else
				{
					throw OutOfMemory();
				}
			}
		}

		// Okay, we are guaranteed enough space now.
		ASSERT(m_curEnd + actualNeededUnits <= m_memoryEnd);
		ASSERT(!(m_used > 0 && m_curEnd <= m_curStart) || !(m_curEnd + actualNeededUnits > m_curStart));

		MemoryUnit* allocateAt = m_curEnd;
		m_curEnd += actualNeededUnits;

		*static_cast<MemoryUnit*>(allocateAt) = unitSize;

		m_used += actualNeededUnits;

		m_itemAddedCond.SignalAll();

		return allocateAt + 1;
	}

	void Free(void* memToFree)
	{
		Mutex::ScopedLock lock(m_monitorLock);

		while(m_used == 0)
		{
			if(m_blocking)
			{
				m_itemAddedCond.Wait(m_monitorLock);
			}
			else
			{
				ASSERT(false);
			}
		}

		ASSERT(m_used > 0);

		if(m_curStart >= m_curEnd && *m_curStart == -1)
		{
			m_curStart = m_memory;
		}

		ASSERT(m_curStart == static_cast<MemoryUnit*>(memToFree) - 1);
		
		int usedUnits = *m_curStart + 1;
		m_curStart += usedUnits; // The first memory unit of an allocation is its size
		m_used -= usedUnits;

		ASSERT(m_used >= 0);

		m_itemRemovedCond.SignalAll();
	}

	bool Empty() const
	{
		return m_used == 0;
	}

	bool Own(void* mem) // Returns whether we own this piece of memory or not
	{
		Mutex::ScopedLock lock(m_monitorLock);

		return m_memory <= mem && mem < m_memoryEnd;
	}
};

#endif
