#ifndef RESOURCECREATOR_H_
#define RESOURCECREATOR_H_

#include <FileFetcher.h>
#include <ThreadPrimitives.h>
#include <ResourceIdentifier.h>
#include <Allocator.h>

#include <boost/bind.hpp>

// The ResourceCreator object manages the 2-step process of A) Loading a file off the disk
// and B) Constructing an object with the loaded data.  This process is encapsulated within
// the ResourceCreator::CreationProcess object.  To begin this process, construct an instance
// of the ResourceCreator::CreationProcess task.  You will be notified of process completion
// via the callback method you pass in to the ResourceCreator::CreationProcess constructor.
// The callback will be called from a seperate thread.
// You may destroy the ResourceCreator::CreationProcess object at any time, however if
// the resource is in the process of being constructed (ie: the load is complete and we now
// have a thread constructing it) the thread calling the destructor will be blocked until
// the object construction has completed (and the callback has been called).

class ResourceFactory;
class ResourceCreatorWorker;
class ResourceCreator
{
	class CreationProcess;
	struct WorkOrder;

	FileFetcher& m_fetcher;

	mutable Mutex m_monitorMutex;

	ThreadSafeAllocator& m_allocator;

	struct WorkerInfo
	{
		ResourceCreatorWorker* worker;
		bool waiting;
		CreationProcess* process;

		WorkerInfo(ResourceCreator& resCreator, int workerId, ThreadSafeAllocator& allocator);
		~WorkerInfo();
	};
	std::vector<WorkerInfo*> m_workers;
	typedef std::map<Thread::ID, int> ThreadIDToWorkerID;
	ThreadIDToWorkerID m_workingThreads;
	

	bool WorkerCanWork() const; // The boolean expression corresponding to the following Condition
	Condition m_workAvailableCond;

	static const int SIMULTANEOUS_WORKERS = 1;
	int NumActiveWorkers() const {return m_numWorkingWorkers - m_numReenteredWorkers;}
	Condition m_idleCond; // Signalled when we are in an idle state

	int m_numReenteredWorkers; // Number of workers that have called back in to the resource creator
	int m_numWorkingWorkers;

	void CreateNewWorker();

	std::list<CreationProcess*> m_pendingProcesses; // Where CreationProcesses sit when they're done loading and waiting to be constructed

	void StartupWorkers();
	void ShutdownWorkers();

	void PrivateScheduleConstruction(CreationProcess* process); // Data is loaded, need to construct it

	void RegisterCreationProcess(CreationProcess* process);
	void UnregisterCreationProcess(CreationProcess* process);

	void MarkProcessWaitedOn(CreationProcess* process, Thread::ID byWho);
	void MarkProcessNotWaitedOn(CreationProcess* process, Thread::ID byWho);


	void FileLoadDoneCB(CreationProcess* process, FileFetcher::FileLoad* fileLoad);
	void HandleFileLoadDone(CreationProcess* process);

	Thread* m_fileLoadDoneThread;
	void RunHandleFileLoadDone();
	typedef std::list<CreationProcess*> FileLoadDoneQueue;
	FileLoadDoneQueue m_fileLoadDoneQueue;
	Mutex m_fileLoadDoneMutex;
	Condition m_fileLoadDoneCond;

	bool m_shuttingDownWorkers;

	bool PrivateIsIdle() const;

	int m_numActiveProcesses;
	void IncrActiveProcesses();
	void DecrActiveProcesses();
public:
	ResourceCreator(FileFetcher& fetcher, ThreadSafeAllocator& allocator);
	~ResourceCreator();

	void WaitForIdle(); // Waits until the ResourceCreator is idle
	bool IsIdle() const; // Are we not currently creating or loading anything?

	// This is basically the public interface of the ResourceCreator class.
	// Creating this object will begin the resource creation process,
	// and destroying it indicates that you no longer care about the creation
	// of the resource (which is especially natural to do if the resource is done loading)
	// You will know when the process is complete by noticing the callback being called.
	// If you destroy the CreationProcess before the callback is made, it will not be
	// called.
	class CreationProcess
	{
		friend class ResourceCreator;
		ResourceCreator& resCreator;
		ResourceFactory& resAlloc;

		typedef void ResourceCreatedSig(CreationProcess* creatingProcess, void* data);
		boost::function<ResourceCreatedSig> resCreatedCB;

		FileFetcher::FileLoad* fileLoad;
		char fileLoadMem[sizeof(FileFetcher::FileLoad)];

		std::list<CreationProcess*>::iterator positionInLine;

		bool inFileLoadDoneQueue;
		Condition notInFileLoadDoneQueueCond;

		Condition constructionDoneCond;

		WorkOrder* workOrder;

		Thread::ID callingThread;

		ResourceIdentifier resId;

		int waitedOnCount;

		bool done;

	public:
		CreationProcess(ResourceCreator& pResCreator, const ResourceIdentifier& pResId, ResourceFactory& pResAlloc, const boost::function<ResourceCreatedSig>& pResCreatedCB, Thread::ID pCallingThread = Thread::GetCurrentID())
			: resCreator(pResCreator)
			, resAlloc(pResAlloc)
			, resCreatedCB(pResCreatedCB)
			, positionInLine(resCreator.m_pendingProcesses.end())
			, inFileLoadDoneQueue(false)
			, workOrder(0)
			, callingThread(pCallingThread)
			, resId(pResId)
			, waitedOnCount(0)
			, done(false)
		{
			ASSERT(!resCreatedCB.empty());

			resCreator.RegisterCreationProcess(this);
		};
		~CreationProcess()
		{
			resCreator.UnregisterCreationProcess(this);

			ASSERT(waitedOnCount == 0);
		}

		void MarkDone(void* data)
		{
			done = true;
			resCreatedCB(this, data);
		}

		const ResourceIdentifier& GetResourceID() const {return resId;}

		// To deal with reentry
		void MarkWaitedOn(Thread::ID byWho) {resCreator.MarkProcessWaitedOn(this, byWho);}
		void MarkNotWaitedOn(Thread::ID byWho) {resCreator.MarkProcessNotWaitedOn(this, byWho);}
	};

	struct WorkOrder
	{
		CreationProcess* creationProcess;

		ResourceFactory& resAlloc;
		FileFetcher::FileLoad* file;

		void* processedData;			// out
		
		WorkOrder(CreationProcess* pCreationProcess, ResourceFactory& pResAlloc, FileFetcher::FileLoad* pFile)
			: creationProcess(pCreationProcess)
			, resAlloc(pResAlloc)
			, file(pFile)
			, processedData(0)
		{}
	};

	WorkOrder* WorkerGetWork(int workerId);
	void WorkerDoneWork(WorkOrder* work, int workerId);
};


class ResourceFactory
{
public:
	virtual void CreateObject(void* mem, std::istream& in) = 0;
	virtual void DestroyObject(void* obj) = 0;
	virtual long ObjectSize() = 0;
};

template <class T>
class TypedResourceFactory : public ResourceFactory
{
public:
	virtual void DestroyObject(void* obj)
	{
		T* typedObj = (T*)obj;
		typedObj->~T();
	}
	virtual long ObjectSize()
	{
		return sizeof(T);
	}
};

#endif
