#include "GraphicsFrame.h"

static const bool USE_SEPERATE_THREAD = false;


bool GraphicsFrame::StateChangeOrder::operator()(const Packet* ls, const Packet* rs) const
{
	// Draw high layers first
	if(ls->GetLayer() > rs->GetLayer()) return true;
	else if(rs->GetLayer() > ls->GetLayer()) return false;

	if(!ls->HasAlpha() && rs->HasAlpha()) return true;
	else if(ls->HasAlpha() && !rs->HasAlpha()) return false;

	if(ls->GetRenderObjectType() < rs->GetRenderObjectType()) return true;
	else if(rs->GetRenderObjectType() < ls->GetRenderObjectType()) return false;

	if(ls->GetMaterial() < rs->GetMaterial()) return true;
	else if(rs->GetMaterial() < ls->GetMaterial()) return false;

	// front to back order if we're drawing opaque, otherwise back to front
	float lsZ = ls->GetModelView().GetVal(2, 3);
	float rsZ = rs->GetModelView().GetVal(2, 3);
	if(ls->HasAlpha())
	{
		ASSERT(rs->HasAlpha());
		// back to front
		if(lsZ < rsZ) return true;
		else if(rsZ < lsZ) return false;
	}
	else
	{
		ASSERT(!rs->HasAlpha());
		// front to back
		if(lsZ > rsZ) return true;
		else if(rsZ > lsZ) return false;
	}
	
	if(ls->GetRenderObjectType() == Packet::RO_TRIMESH)
	{
		if(ls->GetMesh() < rs->GetMesh()) return true;
		else if(rs->GetMesh() < ls->GetMesh()) return false;

		if(ls->GetWireframe() < rs->GetWireframe()) return true;
		else if(rs->GetWireframe() < ls->GetWireframe()) return false;
	}

	return ls < rs;
}

GraphicsFrame::GraphicsFrame(unsigned int maxMatrices, unsigned int maxPackets)
: m_matrices(maxMatrices)
, m_packets(maxPackets)
, m_commandQueue(0)
, m_locked(false)
, m_lightDir(1,0,0,0)
, m_backgroundColour(1,1,1)
{
	if(USE_SEPERATE_THREAD)
	{
		m_commandQueue = new CommandQueue(100, 10000);
	}
}

GraphicsFrame::~GraphicsFrame()
{
	{
		// Wait until the freeze lock is off
		Mutex::ScopedLock lock(m_freezeLock);
	}

	if(USE_SEPERATE_THREAD)
	{
		delete m_commandQueue;
	}
}

void GraphicsFrame::ProcessCommand(AddPacketCommand& cmd)
{
	Mutex::ScopedLock lock(m_sortedPacketsLock);
	m_drawSortedPackets.insert(cmd.packet);
}

void GraphicsFrame::AddPacket(const Packet& packet)
{
	ASSERT(!m_locked);

	m_packets.Push(packet);
	Packet& addPacket = m_packets.Top();

	addPacket.AttachToFrame(*this);
	if(USE_SEPERATE_THREAD)
	{
		ADDCOMMAND(*m_commandQueue, AddPacketCommand(*this, &addPacket));
	}
	else
	{
		Mutex::ScopedLock lock(m_sortedPacketsLock);
		m_drawSortedPackets.insert(&addPacket);
	}
}

void GraphicsFrame::Lock()
{
	if(USE_SEPERATE_THREAD)
	{
		m_commandQueue->Flush();
	}

	m_locked = true;
}
