#include "OpenGLRenderContext.h"

#include "GLee.h"
#include <GL/gl.h>

// The functions in this .cpp file need to be executed by only 1 thread

void OpenGLRenderContext::BeginFrame(const Vector<3>& fillColour, const Matrix<4,4>& projMatrix)
{
	m_curLayer = -1;

	glClearColor(fillColour[0], fillColour[1], fillColour[2],1);
	glClear(GL_COLOR_BUFFER_BIT);

	glMatrixMode(GL_PROJECTION);

	glLoadMatrixf(projMatrix.GetElements());

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glShadeModel(GL_SMOOTH);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);

	glEnable(GL_LIGHTING);
	m_lightsEnabled = true;

	glDisable(GL_FOG);

	glFrontFace(GL_CCW);

	float LightDiffuse[] = {0.9f, 0.9f, 0.9f, 1.0f};
	float LightAmbient[] = {0.3f, 0.3f, 0.3f, 1.0f};
	glEnable(GL_LIGHT0);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, LightDiffuse);
	glLightfv(GL_LIGHT0, GL_AMBIENT, LightAmbient);

	glEnable(GL_NORMALIZE);

	glDisable(GL_SCISSOR_TEST);

	glColor3f(0.0f,1.0f,0.0f);
}

void OpenGLRenderContext::EndFrame()
{
	glFlush();
}

OpenGLRenderContext::OpenGLRenderContext()
: m_curMaterial(0)
, m_prevAlpha(0.0f)
{
}

void OpenGLRenderContext::LoadMaterial(const Material* mat, bool forceReload)
{
	if(mat == m_curMaterial && !forceReload) return;

	if(!m_lightsEnabled)
	{
		glColor4fv(mat->ambientColour.GetElements());
	}
	else
	{	
		glMaterialfv(GL_FRONT, GL_AMBIENT, mat->ambientColour.GetElements());
		glMaterialfv(GL_FRONT, GL_DIFFUSE, mat->diffuseColour.GetElements());
		if(mat->useSpecular)
		{
			glMaterialfv(GL_FRONT, GL_SPECULAR, mat->specularColour.GetElements());
			glMaterialf(GL_FRONT, GL_SHININESS, mat->shininess);
		}
		else
		{
			float zero[4] = {0.0f, 0.0f, 0.0f, 0.0f};
			glMaterialfv(GL_FRONT, GL_SPECULAR, zero);
		}
	}

	m_curMaterial = mat;
}

bool OpenGLRenderContext::DisplayListExists(const TriangleMesh* mesh, bool wireframe) const
{
	const DispListMap& dispMap = wireframe ? m_dispListMapWireframe : m_dispListMap;

	DispListMap::const_iterator found = dispMap.find(mesh);
	return found != dispMap.end();
}

int OpenGLRenderContext::GetDisplayList(const TriangleMesh* mesh, bool wireframe) const
{
	const DispListMap& dispMap = wireframe ? m_dispListMapWireframe : m_dispListMap;

	DispListMap::const_iterator found = dispMap.find(mesh);
	ASSERT(found != dispMap.end());

	return found->second;
}

void OpenGLRenderContext::CreateDisplayList(const TriangleMesh* mesh, bool wireframe)
{
	DispListMap& dispMap = wireframe ? m_dispListMapWireframe : m_dispListMap;

	ASSERT(!DisplayListExists(mesh, wireframe));

	int dispList = glGenLists(1);
	glNewList(dispList, GL_COMPILE);

	glBegin(wireframe ? GL_LINES : GL_TRIANGLES);
	const TriangleMesh::TriangleArray& tris = mesh->GetTriangles();
	const TriangleMesh::VertexArray& verts = mesh->GetVertices();

	for(TriangleMesh::TriangleArray::const_iterator triIter = tris.begin(); triIter != tris.end(); ++triIter)
	{
		const TriangleMesh::Triangle& curTri = *triIter;

		for(int i=0; i<3; ++i)
		{
			glNormal3fv(curTri.GetNormal(i).GetElements());
			glVertex3fv(verts[curTri.GetVertex(i)].GetElements());
		}
	}
	glEnd();

	glEndList();

	dispMap[mesh] = dispList;
}

void OpenGLRenderContext::ApplyAlpha(float alpha)
{
	glBlendColor(1,1,1,alpha);
}

void OpenGLRenderContext::EnableAlpha()
{
	glDepthMask(GL_FALSE);
	glEnable(GL_BLEND);
		
	glBlendFunc(GL_CONSTANT_ALPHA, GL_ONE_MINUS_CONSTANT_ALPHA);
}

void OpenGLRenderContext::DisableAlpha()
{
	glDisable(GL_BLEND);
	glDepthMask(GL_TRUE);
}

void OpenGLRenderContext::ResetLayerInfo()
{
	DisableAlpha();
	m_prevAlpha = 1.1f;
	glClear(GL_DEPTH_BUFFER_BIT);
}

void OpenGLRenderContext::ProcessPacket(const GraphicsFrame::Packet& packet)
{
	if(m_curLayer != packet.GetLayer())
	{
		ResetLayerInfo();
		m_curLayer = packet.GetLayer();
	}

	if(m_lightsEnabled != packet.IsLit())
	{
		if(packet.IsLit())
		{
			glEnable(GL_LIGHTING);
		}
		else
		{
			glDisable(GL_LIGHTING);
		}
		m_lightsEnabled = packet.IsLit();

		LoadMaterial(packet.GetMaterial(), true);
	}

	if(m_curMaterial != packet.GetMaterial())
	{
		LoadMaterial(packet.GetMaterial());
	}

	// Alpha
	if(m_prevAlpha < 1.0f && !packet.HasAlpha())
	{
		DisableAlpha();
	}
	if(m_prevAlpha >= 1.0f && packet.HasAlpha())
	{
		EnableAlpha();
	}
	if(packet.HasAlpha())
	{
		ApplyAlpha(packet.GetAlpha());
	}
	m_prevAlpha = (packet.HasAlpha() ? packet.GetAlpha() : 1.1f);
	

	if(packet.GetRenderObjectType() == GraphicsFrame::Packet::RO_TRIMESH)
	{
		ASSERT(packet.GetMesh() != 0);

		// Get or make the display list for the mesh, and then draw it
		const TriangleMesh* drawMesh = packet.GetMesh();
		if(!DisplayListExists(drawMesh, packet.GetWireframe()))
		{
			CreateDisplayList(drawMesh, packet.GetWireframe());
		}
		int dispList = GetDisplayList(drawMesh, packet.GetWireframe());

		// Draw it!
		glLoadMatrixf(packet.GetModelView().GetElements());

		if(packet.GetWireframe())
		{
			glLineWidth(1.0f);
		}

		glCallList(dispList);
	}
	else if(packet.GetRenderObjectType() == GraphicsFrame::Packet::RO_LINE)
	{
		ASSERT(packet.GetLine() != 0);

		glLineWidth(packet.GetLineThickness());

		// Draw it!
		glLoadMatrixf(packet.GetModelView().GetElements());

		glBegin(GL_LINES);
		
		glVertex3fv(packet.GetLine()->a.Trim().GetElements());
		glVertex3fv(packet.GetLine()->b.Trim().GetElements());

		glEnd();		
	}
	else
	{
		ASSERT(false);
	}
}

void OpenGLRenderContext::RenderGraphicsFrame(GraphicsFrame& frame, const Matrix<4,4>& projMatrix, const RenderOptions& options)
{
	BeginFrame(frame.GetBackgroundColour(), projMatrix);
	
	glLightfv(GL_LIGHT0, GL_POSITION, frame.GetLightDir().Normalized().GetElements());

	// Iterate through the graphics frame drawing each packet
	{
		GraphicsFrame::SortedPacketSet packets = frame.GetPackets();

		for(GraphicsFrame::SortedPacketSet::const_iterator iter = packets.begin();
			iter != packets.end();
			++iter)
		{
			ProcessPacket(**iter);
		}
	}

	EndFrame();
}

void OpenGLRenderContext::SetViewportSize(int width, int height)
{
	glViewport(0,0, width, height);
	glScissor(0,0, width, height);
}
