#include "ResourceCreator.h"
#include "ResourceCreatorWorker.h"

#include <iostream>

ResourceCreator::ResourceCreator(FileFetcher& fetcher, ThreadSafeAllocator& allocator)
: m_fetcher(fetcher)
, m_numReenteredWorkers(0)
, m_numWorkingWorkers(0)
, m_allocator(allocator)
, m_shuttingDownWorkers(false)
, m_numActiveProcesses(0)
{
	m_fileLoadDoneThread = new Thread(boost::bind(&ResourceCreator::RunHandleFileLoadDone, this));

	StartupWorkers();
}

ResourceCreator::~ResourceCreator()
{
	ASSERT(m_numActiveProcesses == 0);
	ShutdownWorkers();

	{
		Mutex::ScopedLock lock(m_fileLoadDoneMutex);
		m_fileLoadDoneQueue.push_back(0);
		m_fileLoadDoneCond.SignalAll();
	}

	delete m_fileLoadDoneThread;
}

ResourceCreator::WorkerInfo::WorkerInfo(ResourceCreator& resCreator, int workerId, ThreadSafeAllocator& allocator)
	: worker(new ResourceCreatorWorker(resCreator, workerId, allocator))
	, waiting(false)
{
}
ResourceCreator::WorkerInfo::~WorkerInfo()
{
	delete worker;
}

void ResourceCreator::RegisterCreationProcess(CreationProcess* process)
{
	Mutex::ScopedLock lock(m_monitorMutex);
	
	if(process->resId.IsRelative())
	{
		// If they gave us a relative path, make it an absolute path before proceeding

		ThreadIDToWorkerID::iterator found = m_workingThreads.find(process->callingThread);
		if(found != m_workingThreads.end())
		{
			// This creation process was kicked off by a worker, prepend the path of the file the 
			// worker is loading to this filename
			process->resId = m_workers[found->second]->process->resId.Append(process->resId);
		}
	}

	process->fileLoad = new FileFetcher::FileLoad(m_fetcher, process->resId, boost::bind(&ResourceCreator::FileLoadDoneCB, this, process, _1));

	++m_numActiveProcesses;
}

void ResourceCreator::MarkProcessWaitedOn(CreationProcess* process, Thread::ID byWho)
{
	Mutex::ScopedLock lock(m_monitorMutex);

	// Check if this creation process was initiated by one of our workers... if so, update the reentry count
	ThreadIDToWorkerID::iterator found = m_workingThreads.find(byWho);
	if(found != m_workingThreads.end())
	{
		// Uh ouh, a reentry!
		ASSERT(found->second < m_workers.size());
		WorkerInfo& reenteredWorker = *m_workers[found->second];

		ASSERT(!reenteredWorker.waiting); // How can they be waiting twice?
		reenteredWorker.waiting = true;

		++m_numReenteredWorkers;

		if(NumActiveWorkers() < SIMULTANEOUS_WORKERS)
		{
			if(m_numWorkingWorkers == m_workers.size())
			{
				// Need more workers!
				CreateNewWorker();
			}
			else if(!m_pendingProcesses.empty())
			{
				// Wake up an idle worker
				m_workAvailableCond.SignalAll();
			}
		}
	}

	++process->waitedOnCount;
}
void ResourceCreator::MarkProcessNotWaitedOn(CreationProcess* process, Thread::ID byWho)
{
	Mutex::ScopedLock lock(m_monitorMutex);

	ThreadIDToWorkerID::iterator found = m_workingThreads.find(byWho);
	if(found != m_workingThreads.end())
	{
		--m_numReenteredWorkers;

		ASSERT(m_workers[found->second]->waiting);
		m_workers[found->second]->waiting = false;
	}

	--process->waitedOnCount;
}

void ResourceCreator::RunHandleFileLoadDone()
{
	while(true)
	{
		CreationProcess* next;

		{
			Mutex::ScopedLock lock(m_fileLoadDoneMutex);
			if(m_fileLoadDoneQueue.empty())
			{
				m_fileLoadDoneCond.Wait(m_fileLoadDoneMutex);
			}
			ASSERT(!m_fileLoadDoneQueue.empty());

			next = m_fileLoadDoneQueue.front();
			m_fileLoadDoneQueue.pop_front();
		}

		if(next == 0) break; // quit command

		// otherwise, handle it
		HandleFileLoadDone(next);

		{
			Mutex::ScopedLock lock(m_fileLoadDoneMutex);
			ASSERT(next->inFileLoadDoneQueue);
			next->inFileLoadDoneQueue = false;
			next->notInFileLoadDoneQueueCond.SignalAll();
		}
	}
}

void ResourceCreator::HandleFileLoadDone(CreationProcess* process)
{
	Mutex::ScopedLock lock(m_monitorMutex); // suuuuch a stupid idea.  

	// Okay, this is where the ResourceCreator actually starts
	// doing something (up until now the file was just getting fetched)

	ASSERT(process->fileLoad != 0);
	if(process->fileLoad->Error())
	{
		// we have no more use for the file load if there was an error
		delete process->fileLoad;
		process->fileLoad = 0;

		// Propogate the file load error to a CreationProcess error
		process->MarkDone(0);
		DecrActiveProcesses();

		std::cerr << "Error opening resource \"" << process->resId.GetFilepath().string() << "\"." << std::endl;

		return;
	}
	else
	{
		// Otherwise, look in to getting this sweet raw data processed
		PrivateScheduleConstruction(process);
	}
}

void ResourceCreator::FileLoadDoneCB(CreationProcess* process, FileFetcher::FileLoad* fileLoad)
{
	Mutex::ScopedLock lock(m_fileLoadDoneMutex);
	m_fileLoadDoneQueue.push_back(process);
	process->inFileLoadDoneQueue = true;
	m_fileLoadDoneCond.SignalAll();
}

bool ResourceCreator::WorkerCanWork() const
{
	return !m_pendingProcesses.empty() && NumActiveWorkers() < SIMULTANEOUS_WORKERS;
}

void ResourceCreator::WaitForIdle()
{
	Mutex::ScopedLock lock(m_monitorMutex);

	while(!PrivateIsIdle())
	{
		m_idleCond.Wait(m_monitorMutex);
	}
}
bool ResourceCreator::IsIdle() const
{
	Mutex::ScopedLock lock(m_monitorMutex);
	return PrivateIsIdle();
}
bool ResourceCreator::PrivateIsIdle() const
{
	return m_numActiveProcesses == 0;
}

void ResourceCreator::PrivateScheduleConstruction(CreationProcess* process)
{
	ThreadIDToWorkerID::iterator found;
	std::list<CreationProcess*>::iterator at;
	
	if(process != 0 && process->waitedOnCount > 0)
	{
		// Looks like we're probably scheduled by another worker...
		// Bump our priority so the original worker can resume faster
		at = m_pendingProcesses.insert(m_pendingProcesses.begin(), process);
	}
	else
	{
		at = m_pendingProcesses.insert(m_pendingProcesses.end(), process);
	}

	if(process != 0)
	{
		process->positionInLine = at;
	}

	// Wake up any threads waiting for work
	if(WorkerCanWork())
	{
		if(m_numWorkingWorkers == m_workers.size())
		{
			CreateNewWorker();
		}
		else
		{
			m_workAvailableCond.SignalAll();
		}
	}
}

void ResourceCreator::CreateNewWorker()
{
	m_workers.push_back(new WorkerInfo(*this, m_workers.size(), m_allocator));

	if(m_shuttingDownWorkers)
	{
		PrivateScheduleConstruction(0);
	}
}

ResourceCreator::WorkOrder* ResourceCreator::WorkerGetWork(int workerId)
{
	Mutex::ScopedLock lock(m_monitorMutex);

	ASSERT(workerId < m_workers.size());

	// We want to keep only one worker active at a time
	while(!WorkerCanWork())
	{
		m_workAvailableCond.Wait(m_monitorMutex);
	}

	// Return the next work item
	CreationProcess* nextProcess = m_pendingProcesses.front();
	m_pendingProcesses.pop_front();

	if(nextProcess == 0) return 0; // Quit command

	nextProcess->positionInLine = m_pendingProcesses.end();

	// Update our worker stats
	++m_numWorkingWorkers;
	ASSERT(m_workingThreads.find(m_workers[workerId]->worker->GetThreadID()) == m_workingThreads.end());
	m_workingThreads[m_workers[workerId]->worker->GetThreadID()] = workerId;

	m_workers[workerId]->process = nextProcess;

	nextProcess->workOrder = new WorkOrder(nextProcess, nextProcess->resAlloc, nextProcess->fileLoad);
	return nextProcess->workOrder;
}

void ResourceCreator::WorkerDoneWork(WorkOrder* work, int workerId)
{
	Mutex::ScopedLock lock(m_monitorMutex);

	ASSERT(workerId < m_workers.size());
	ASSERT(m_numWorkingWorkers > 0);
	--m_numWorkingWorkers;
	ThreadIDToWorkerID::iterator found = m_workingThreads.find(m_workers[workerId]->worker->GetThreadID());
	ASSERT(found != m_workingThreads.end()); // This definitely WAS a working thread
	m_workingThreads.erase(found); // but not anymore! 

	ASSERT(!m_workers[workerId]->waiting);

	// Don't need the file load anymore
	delete work->creationProcess->fileLoad;
	work->creationProcess->fileLoad = 0;


	// They're reporting in with completed work
	// We're done!  Fire callback! Yeah! (note, work->processedData could be 0 here if there was an error)
	work->creationProcess->MarkDone(work->processedData);
	DecrActiveProcesses();

	ASSERT(work->creationProcess->workOrder != 0);
	work->creationProcess->workOrder = 0;

	work->creationProcess->constructionDoneCond.SignalAll();

	delete work;
}

void ResourceCreator::UnregisterCreationProcess(CreationProcess* process)
{
	{
		Mutex::ScopedLock lock(m_monitorMutex);
		if(process->fileLoad != 0)
		{
			process->fileLoad->StopLoadProcess();
		}
	}

	{
		Mutex::ScopedLock lock(m_fileLoadDoneMutex);
		if(process->inFileLoadDoneQueue)
		{
			process->notInFileLoadDoneQueueCond.Wait(m_fileLoadDoneMutex);
		}
		ASSERT(!process->inFileLoadDoneQueue);
	}

	Mutex::ScopedLock lock(m_monitorMutex);

	if(process->positionInLine != m_pendingProcesses.end())
	{
		// It's waiting in line, remove it from the line
		m_pendingProcesses.erase(process->positionInLine);
		process->positionInLine = m_pendingProcesses.end();
		DecrActiveProcesses();
	}
	else if(process->workOrder != 0)
	{
		// If we started constructing it, we have to wait until it finishes
		process->constructionDoneCond.Wait(m_monitorMutex);
	}

	if(process->fileLoad != 0)
	{
		delete process->fileLoad;
		process->fileLoad = 0;
	}

	if(!process->done)
	{
		DecrActiveProcesses();
	}
}


void ResourceCreator::StartupWorkers()
{
	// Nothing to do here, workers will get started up on demand
}

void ResourceCreator::ShutdownWorkers()
{
	{
		Mutex::ScopedLock lock(m_monitorMutex);

		for(int i=0; i<m_workers.size(); ++i)
		{
			PrivateScheduleConstruction(0);
		}

		m_shuttingDownWorkers = true;
	}

	while(!m_workers.empty())
	{
		delete m_workers.back();
		m_workers.pop_back();
	}

	ASSERT(m_numWorkingWorkers == 0);
	ASSERT(m_numReenteredWorkers == 0);
}


void ResourceCreator::IncrActiveProcesses()
{
	ASSERT(m_monitorMutex.IsLocked());
	++m_numActiveProcesses;
}

void ResourceCreator::DecrActiveProcesses()
{
	ASSERT(m_monitorMutex.IsLocked());

	--m_numActiveProcesses;
	ASSERT(m_numActiveProcesses >= 0);

	if(m_numActiveProcesses == 0)
	{
		m_idleCond.SignalAll();
	}
}

