#include "ResourceManager.h"


ResourceManager::ResourceManager(Allocator& alloc, long maxCapacity, FileFetcher& fetcher)
: m_alloc(alloc)
, m_maxCapacity(maxCapacity)
, m_curMemUsed(0)
, m_curNonFreeableMemUsed(0)
{
	m_resCreator = new ResourceCreator(fetcher, *this);

	m_resCreatedHandlerThread = new Thread(boost::bind(&ResourceManager::RunResCreatedHandlingThread, this));
	m_resDestroyerThread = new Thread(boost::bind(&ResourceManager::RunResourceDestroyer, this));
}

ResourceManager::~ResourceManager()
{
	// Go through the allocations and make sure they all finish up
	// We put this cleanup code in a loop because it's possible that a resource
	// has a reference to another resource, and so the acquisition count will 
	// be greater than 0 until we free the original reference from the freeable list.
	int remainingResources = -1;
	while(remainingResources != 0)
	{
		int curRemainingResources = 0;

		while(!m_resCreator->IsIdle())
		{
			m_resCreator->WaitForIdle();

			// Wait for the resource creator to create any newly constructed resources
			{
				Mutex::ScopedLock lock(m_resCreatedMutex);
				if(!m_resCreatedQueue.empty())
				{
					m_noResToCreateCond.Wait(m_resCreatedMutex);
				}
				ASSERT(m_resCreatedQueue.empty()); // We should always only have to wait at most once
			}
		}

		{
			Mutex::ScopedLock lock(m_monitorMutex);
		

			for(AcquisitionMap::iterator iter = m_acquisitions.begin(); iter != m_acquisitions.end(); ++iter)
			{
				if(iter->second->acquisitionCount > 0) ++curRemainingResources;
			}
			 
			// Start killing the freeable memory, which should free up some of the still acquired resources
			while(!m_freeableList.empty())
			{
				Free(m_freeableList.front());
			}
		}

		// Wait for the resource destroyer to destroy everything
		{
			Mutex::ScopedLock lock(m_resToDestroyMutex);
			if(!m_destroyList.empty())
			{
				m_noResToDestroyCond.Wait(m_resToDestroyMutex);
			}
			ASSERT(m_destroyList.empty());
		}

		ASSERTMSG(curRemainingResources != remainingResources, "You still hold an acquisition to some resources.");
		remainingResources = curRemainingResources;
	}


	{
		Mutex::ScopedLock lock(m_resCreatedMutex);
		m_resCreatedQueue.push_back(0);
		m_resCreatedCond.SignalAll();
	}
	delete m_resCreatedHandlerThread;

	{
		Mutex::ScopedLock lock(m_resToDestroyMutex);
		m_destroyList.push_back(0);
		m_resToDestroyCond.SignalAll();
	}
	delete m_resDestroyerThread;

	ASSERT(m_curNonFreeableMemUsed == 0);


	ASSERT(m_curMemUsed == 0);

	delete m_resCreator;
}



void* ResourceManager::Allocate(long bytes)
{
	Mutex::ScopedLock lock(m_monitorMutex);

	ASSERTMSG(bytes + m_curNonFreeableMemUsed <= m_maxCapacity, "The resource allocator has run out of memory!");

	// Free the freeable allocations until we have enough room
	while(bytes + m_curMemUsed > m_maxCapacity)
	{
		ASSERT(!m_freeableList.empty());

		Free(m_freeableList.front());
	}

	// Create our allocation block
	void* allocated = m_alloc.Allocate(bytes);

	AllocationInfo* newInfo = new AllocationInfo();
	newInfo->allocationSize = bytes;
	newInfo->freeablePosition = m_freeableList.end();
	newInfo->resAcquisition = m_acquisitions.end();

	m_allocations[allocated] = newInfo;

	m_curNonFreeableMemUsed += bytes;
	m_curMemUsed += bytes;

	return allocated;
}

void ResourceManager::RunResourceDestroyer()
{
	while(true)
	{
		ResToDestroyInfo* nextInfo;

		{
			Mutex::ScopedLock lock(m_resToDestroyMutex);

			if(m_destroyList.empty())
			{
				m_noResToDestroyCond.SignalAll();
				m_resToDestroyCond.Wait(m_resToDestroyMutex);
			}
			ASSERT(!m_destroyList.empty());

			nextInfo = m_destroyList.front();
		}

		if(nextInfo == 0)
		{
			Mutex::ScopedLock lock(m_resToDestroyMutex);
			m_destroyList.pop_front();
			break;
		}

		// Destroy it!
		nextInfo->resFactory->DestroyObject(nextInfo->data);

		{
			Mutex::ScopedLock lock(m_monitorMutex);
			m_alloc.Free(nextInfo->data);
		}

		delete nextInfo;

		{
			Mutex::ScopedLock lock(m_resToDestroyMutex);
			m_destroyList.pop_front();
		}
	}
}

void ResourceManager::Free(AllocationMap::iterator iter)
{
	// Destruct this bitch (it shouldn't be on the freeable list if it wasn't fully constructed)
	m_curMemUsed -= iter->second->allocationSize; // Get our memory back!
	ASSERT(m_curMemUsed >= 0);

	if(iter->second->resAcquisition != m_acquisitions.end())
	{
		ASSERT(iter->second->resAcquisition->second->resFactory->ObjectSize() == iter->second->allocationSize);

		{
			// Ship the resource off to be destroyed by the resource destroyer
			Mutex::ScopedLock lock(m_resToDestroyMutex);
			ResToDestroyInfo* resDestroyInfo = new ResToDestroyInfo();
			resDestroyInfo->data = iter->first;
			resDestroyInfo->resFactory = iter->second->resAcquisition->second->resFactory;

			m_destroyList.push_back(resDestroyInfo);
			m_resToDestroyCond.SignalAll();
		}

		ASSERT(iter->second->freeablePosition != m_freeableList.end()); // Assert that it is indeed freeable
		m_freeableList.erase(iter->second->freeablePosition); // Remove it from the freeable list

		// Now cleanup the acquisition as well (which should not be being referenced)
		if(iter->second->resAcquisition != m_acquisitions.end())
		{
			ASSERT(iter->second->resAcquisition->second->acquisitionCount == 0);

			// Get rid of the acquisition metadata
			delete iter->second->resAcquisition->second;

			// Remove it from the acquisition list
			m_acquisitions.erase(iter->second->resAcquisition);
		}
	}
	else
	{
		// This object was never constructed, so just free it and forget about it
		m_alloc.Free(iter->first);
	}

	// Delete the allocation metadata
	delete iter->second;

	m_allocations.erase(iter); // We never existed...
}

void ResourceManager::Free(void* mem)
{
	Mutex::ScopedLock lock(m_monitorMutex);

	AllocationMap::iterator found = m_allocations.find(mem);
	ASSERT(found != m_allocations.end());

	Free(found);
}


void ResourceManager::SetFreeableFlag(AllocationMap::iterator& allocIter, bool freeable)
{
	ASSERT(allocIter->second->resAcquisition != m_acquisitions.end()); // This needs to be set before we can adjust the freeability status

	if(freeable)
	{
		// Make sure it wasn't previously freeable
		ASSERT(allocIter->second->freeablePosition == m_freeableList.end());

		allocIter->second->freeablePosition = m_freeableList.insert(m_freeableList.end(), allocIter);

		// Gain some non freeable capacity back
		m_curNonFreeableMemUsed -= allocIter->second->resAcquisition->second->resFactory->ObjectSize();
	}
	else // Make it non-freeable
	{
		// Make sure it was previously freeable
		ASSERT(allocIter->second->freeablePosition != m_freeableList.end());

		// Take it out of the freeable list
		m_freeableList.erase(allocIter->second->freeablePosition);
		allocIter->second->freeablePosition = m_freeableList.end();

		// Account for the reduced non-freeable capacity
		m_curNonFreeableMemUsed += allocIter->second->resAcquisition->second->resFactory->ObjectSize();
	}
}

void ResourceManager::RunResCreatedHandlingThread()
{
	Mutex::ScopedLock lock(m_resCreatedMutex);

	while(true)
	{
		while(m_resCreatedQueue.empty())
		{
			m_resCreatedCond.Wait(m_resCreatedMutex);
		}

		ResCreatedInfo* nextInfo = m_resCreatedQueue.front();
		m_resCreatedQueue.pop_front();

		if(nextInfo == 0) break; // quit command

		ProcessResourceCreated(nextInfo->creationProcess, nextInfo->data);

		delete nextInfo;

		if(m_resCreatedQueue.empty()) m_noResToCreateCond.SignalAll();
	}
}

void ResourceManager::ProcessResourceCreated(ResourceCreator::CreationProcess* creationProcess, void* data)
{
	Mutex::ScopedLock lock(m_monitorMutex);

	AcquisitionMap::iterator acqFound = m_acquisitions.find(creationProcess->GetResourceID());
	ASSERT(acqFound != m_acquisitions.end());
	ASSERT(acqFound->second->creationProcess.GetPtr() == creationProcess);

	if(data == 0)
	{
		// Error... bummer
		acqFound->second->error = true;

		if(acqFound->second->acquisitionCount == 0)
		{
			// Kill it, it's okay to do this now if there was an error
			delete acqFound->second;
			m_acquisitions.erase(acqFound);
		}
		else
		{
			// let everyone know about the disastrous news
			acqFound->second->resourceAvailableCond.SignalAll();
		}
		return;
	}

	// Okay great, let's link up the allocation with the resource acquisition now
	AllocationMap::iterator allocFound = m_allocations.find(data);
	ASSERT(allocFound != m_allocations.end());

	allocFound->second->resAcquisition = acqFound;
	acqFound->second->allocInfo = allocFound;

	// If the resource is no longer being acquired though, put it straight on to the
	// freeable list
	if(acqFound->second->acquisitionCount == 0)
	{
		SetFreeableFlag(allocFound, true);
	}
	else
	{
		// Let everyone know it's available now
		acqFound->second->resourceAvailableCond.SignalAll();
	}
}

void ResourceManager::HandleResourceCreated(ResourceCreator::CreationProcess* creationProcess, void* data)
{
	Mutex::ScopedLock lock(m_resCreatedMutex);

	ResCreatedInfo* info = new ResCreatedInfo();
	info->creationProcess = creationProcess;
	info->data = data;

	m_resCreatedQueue.push_back(info);
	m_resCreatedCond.SignalAll();
}

ResourceManager::AcquisitionMap::iterator ResourceManager::Acquire(const ResourceIdentifier& resId, std::auto_ptr<ResourceFactory> resFactory)
{
	Mutex::ScopedLock lock(m_monitorMutex);

	// Either add the resource acquisition to the set if it doesn't exist, or
	// increment the reference count if it does.
	AcquisitionMap::iterator found = m_acquisitions.find(resId);
	if(found == m_acquisitions.end())
	{
		// Need to add a new acquisition here
		ResourceAcquisition* newAcquisition = new ResourceAcquisition();
		newAcquisition->resFactory = resFactory;
		newAcquisition->acquisitionCount = 1;
		newAcquisition->allocInfo = m_allocations.end();
		newAcquisition->error = false;

		// Start the creation process
		new(newAcquisition->creationProcess) ResourceCreator::CreationProcess(*m_resCreator, resId, *(newAcquisition->resFactory.get()), boost::bind(&ResourceManager::HandleResourceCreated, this, _1, _2));

		// We add this with the newly created CreationProcess' resId instead of the passed in one because
		// the CreationProcess might modify the resource ID if it was relative to another resource that
		// is currently being loaded (identified by whether or not the current thread is a resource constructor worker or not)
		std::pair<AcquisitionMap::iterator, bool> added = m_acquisitions.insert(std::make_pair(newAcquisition->creationProcess->GetResourceID(), newAcquisition));
		ASSERT(added.second);

		found = added.first;
	}
	else
	{
		if(found->second->acquisitionCount == 0 && found->second->allocInfo != m_allocations.end())
		{
			// Yikes, it was in the freeable list, take it out, it's referenced again!
			SetFreeableFlag(found->second->allocInfo, false);
		}

		// Update the reference count
		++found->second->acquisitionCount;
	}

	return found;
}

ResourceManager::AcquisitionMap::iterator ResourceManager::Reacquire(const AcquisitionMap::iterator& iter)
{
	Mutex::ScopedLock lock(m_monitorMutex);

	ASSERT(iter->second->acquisitionCount > 0);

	++iter->second->acquisitionCount;

	return iter;
}

void ResourceManager::Release(AcquisitionMap::iterator& iter)
{
	Mutex::ScopedLock lock(m_monitorMutex);

	ASSERT(iter != m_acquisitions.end());

	--iter->second->acquisitionCount;

	if(iter->second->acquisitionCount == 0 && iter->second->allocInfo != m_allocations.end())
	{
		if(iter->second->error)
		{
			// Just delete it, there was an error so no point in keeping this around
			delete iter->second;
			m_acquisitions.erase(iter);
		}
		else
		{
			// Add this to the freeable list because nobody is referencing it
			SetFreeableFlag(iter->second->allocInfo, true);
		}
	}

	iter = m_acquisitions.end();
}

ResourceManager::AcquisitionStatus ResourceManager::GetAcquisitionStatus(const AcquisitionMap::iterator& iter) const
{
	Mutex::ScopedLock lock(m_monitorMutex);

	// The acquisition is acquired only if the acquisition structure has been linked to the allocation
	if(!iter->second->error && iter->second->allocInfo == m_allocations.end()) return ACQUIRING;

	return ACQUIRED;

}

void* ResourceManager::GetAcquiredResourceData(const AcquisitionMap::iterator& iter) const
{
	Mutex::ScopedLock lock(m_monitorMutex);

	if(!iter->second->error && iter->second->allocInfo == m_allocations.end())
	{
		iter->second->creationProcess->MarkWaitedOn(Thread::GetCurrentID());
		
		iter->second->resourceAvailableCond.Wait(m_monitorMutex); // Wait for the resource to become available

		iter->second->creationProcess->MarkNotWaitedOn(Thread::GetCurrentID());
	}

	if(iter->second->error) return 0;

	ASSERT(iter->second->allocInfo != m_allocations.end());
	return iter->second->allocInfo->first;
}

