#include <Algebra.h>
#include <R3Utils.h>

#include <iostream>

static bool DoubleEquals(double a, double b)
{
	const float EPSILON = 0.00001;
	return abs(a - b) < EPSILON;
}

static bool ComplexEquals(const std::complex<double>& ls, const std::complex<double>& rs)
{
	return DoubleEquals(real(ls), real(rs))
		&& DoubleEquals(imag(ls), imag(rs));
}

static void CheckComplexSolutions(const Tuple<std::complex<double>, 3>& sols, const std::complex<double>& s1, const std::complex<double>& s2, const std::complex<double>& s3)
{
	ASSERT(ComplexEquals(sols[0], s1)
		|| ComplexEquals(sols[0], s2)
		|| ComplexEquals(sols[0], s3));
	ASSERT(ComplexEquals(sols[1], s1)
		|| ComplexEquals(sols[1], s2)
		|| ComplexEquals(sols[1], s3));
	ASSERT(ComplexEquals(sols[2], s1)
		|| ComplexEquals(sols[2], s2)
		|| ComplexEquals(sols[2], s3));
}

int main(int argc, char** args)
{
	{
		Vector<3> testVec1(1,2,3);
		Vector<3> testVec2(4,5,6);

		Vector<3> result = testVec1 + testVec2;
		ASSERT(result[0] == 5);
		ASSERT(result[1] == 7);
		ASSERT(result[2] == 9);

		result = testVec1 - testVec2;
		ASSERT(result[0] == -3);
		ASSERT(result[1] == -3);
		ASSERT(result[2] == -3);

		result = testVec1 * 2;
		ASSERT(result[0] == 2);
		ASSERT(result[1] == 4);
		ASSERT(result[2] == 6);

		ASSERT(testVec1.Trim()[0] == 1);
		ASSERT(testVec1.Trim()[1] == 2);

		float sResult = Dot(testVec1, testVec2);
		ASSERT(sResult == 32);
	}

	{
		Vector<4> testVec1(1,2,3,4);
		Vector<4> testVec2(5,6,7,8);

		Vector<4> result = testVec1 + testVec2;
		ASSERT(result[0] == 6);
		ASSERT(result[1] == 8);
		ASSERT(result[2] == 10);
		ASSERT(result[3] == 12);

		result = testVec1 - testVec2;
		ASSERT(result[0] == -4);
		ASSERT(result[1] == -4);
		ASSERT(result[2] == -4);
		ASSERT(result[3] == -4);

		result = testVec1 * 2;
		ASSERT(result[0] == 2);
		ASSERT(result[1] == 4);
		ASSERT(result[2] == 6);
		ASSERT(result[3] == 8);

		ASSERT(testVec1.Trim()[0] == 1);
		ASSERT(testVec1.Trim()[1] == 2);
		ASSERT(testVec1.Trim()[2] == 3);

		float sResult = Dot(testVec1, testVec2);
		ASSERT(sResult == 70);
	}

	{
		Vector<3> testVec1(1,2,3);
		Vector<3> testVec2(testVec1);

		ASSERT(testVec2[0] == 1);
		ASSERT(testVec2[1] == 2);
		ASSERT(testVec2[2] == 3);
	}
	{
		Vector<4> testVec1(1,2,3,4);
		Vector<4> testVec2(testVec1);

		ASSERT(testVec2[0] == 1);
		ASSERT(testVec2[1] == 2);
		ASSERT(testVec2[2] == 3);
		ASSERT(testVec2[3] == 4);
	}

	{
		float vals[] = {1,2,3};
		Vector<3> testVec(vals);
		
		ASSERT(testVec[0] == 1);
		ASSERT(testVec[1] == 2);
		ASSERT(testVec[2] == 3);
	}
	{
		float vals[] = {1,2,3,4};
		Vector<4> testVec(vals);
		
		ASSERT(testVec[0] == 1);
		ASSERT(testVec[1] == 2);
		ASSERT(testVec[2] == 3);
		ASSERT(testVec[3] == 4);
	}

	{
		// solve a^3 -6a^2 +11a -6
		Tuple<std::complex<double>, 3> solutions = SolveCubic(1.0, -6.0, 11.0, -6.0);

		CheckComplexSolutions(solutions, 1.0, 2.0, 3.0);
	}
	{
		// solve a^3 -6a^2 +11a -6
		Tuple<std::complex<double>, 3> solutions = SolveCubic(-6.0, 11.0, -6.0);

		CheckComplexSolutions(solutions, 1.0, 2.0, 3.0);
	}
	{
		// solve a^3 +5a^2 +(-3+4i)a +(-15 +20i)
		Tuple<std::complex<double>, 3> solutions = SolveCubic(5.0, std::complex<double>(-3.0, 4.0), std::complex<double>(-15.0, 20.0));

		CheckComplexSolutions(solutions, std::complex<double>(2.0, -1.0), std::complex<double>(-2.0, 1.0), -5.0);
	}

	{
		Matrix<3,3> id(Matrix<3,3>::Identity);
		Tuple<std::complex<double>, 3> charPoly = GetCharPoly(id);
		
		ASSERT(ComplexEquals(charPoly[0], -3.0));
		ASSERT(ComplexEquals(charPoly[1], 3.0));
		ASSERT(ComplexEquals(charPoly[2], -1.0));
	}

	{
		// This is a test case introduced because things fucked up on me and this
		// failed before.
		Matrix<3,3> testMat;

		testMat.GetVal(0,0) = 0.73572922;
		testMat.GetVal(1,0) = 0.57149285;
		testMat.GetVal(2,0) = -0.36345360;
		testMat.GetVal(0,1) = 0.27449787;
		testMat.GetVal(1,1) = 0.23897362;
		testMat.GetVal(2,1) = 0.93141961;
		testMat.GetVal(0,2) = 0.61915553;
		testMat.GetVal(1,2) = -0.78503990;
		testMat.GetVal(2,2) = 0.018946346;

		Tuple<std::complex<double>, 3> charPoly = GetCharPoly(testMat);

		ASSERT(ComplexEquals(charPoly[0], -0.993649186));
		ASSERT(ComplexEquals(charPoly[1], 0.993649226));
		ASSERT(ComplexEquals(charPoly[2], -1.000000111));
	}

	{
		Matrix<4,4> trans = MakeUniScaleMatrix(4.0)*MakeRotationMatrix(M_PI/4.0, 'z');
		Matrix<3,3> mat;
		mat.SetColumn(0, trans.GetColumn(0).Trim());
		mat.SetColumn(1, trans.GetColumn(1).Trim());
		mat.SetColumn(2, trans.GetColumn(2).Trim());
		Tuple<std::complex<double>, 3> charPoly = GetCharPoly(mat);
		
		Tuple<std::complex<double>, 3> eigens = SolveCubic(charPoly[0], charPoly[1], charPoly[2]);

		ASSERT(DoubleEquals(abs(eigens[0]), 4.0));
		ASSERT(DoubleEquals(abs(eigens[1]), 4.0));
		ASSERT(DoubleEquals(abs(eigens[2]), 4.0));
	}



	std::cout << "Success." << std::endl;

	return 0;
}
