#ifndef WORLDINSTANCE_H_
#define WORLDINSTANCE_H_

#include "WorldDefinition.h"
#include <TemplateList.h>

#include <ThreadPrimitives.h>

struct SymbolInstance
{
	Symbol* m_sym;
	Production* m_prod;

	SymbolInstance* m_parent;
	
	int m_numChildren;
	bool m_hasChildren;
	
	SymbolInstance* m_children;
	
	bool m_mustReinstance; // Set for making sure the old reference instance always make it in to new pools
	
	SymbolInstance()
	: m_sym(0)
	, m_prod(0)
	, m_parent(0)
	, m_hasChildren(false)
	, m_children(0)
	, m_mustReinstance(false)
	{};
	
	const Production::ChildDescriptor* GetMyChildDescriptorFromParent();
};

class SymbolInstancePool
{
	class BasePool
	{
		SymbolInstance* m_poolMemory;
		SymbolInstance* m_curPosition;
		int m_poolSize;
	
		Mutex m_monitorLock;

		public:
		BasePool(int poolSize);
		~BasePool();
	
		SymbolInstance* MakeInstance();
		SymbolInstance* MakeInstanceArray(int numInstances);
	
		bool Contains(SymbolInstance* sym) const
		{
			return sym >= m_poolMemory && sym < m_curPosition;
		}

		void Reset()
		{
			m_curPosition = m_poolMemory;
		}
	};
	
	BasePool m_mainPool;
	BasePool m_reservePool;
	
	public:
		
	SymbolInstancePool(int poolSize, int reserveSize)
	: m_mainPool(poolSize)
	, m_reservePool(reserveSize)
	{};
	
	SymbolInstance* MakeInstance(bool useReserve = false)
	{
		return (useReserve ? m_reservePool : m_mainPool).MakeInstance();
	}

	SymbolInstance* MakeInstanceArray(int numInstances, bool useReserve = false)
	{
		return (useReserve ? m_reservePool : m_mainPool).MakeInstanceArray(numInstances);
	}
	bool Contains(SymbolInstance* sym) const
	{
		return m_mainPool.Contains(sym) || m_reservePool.Contains(sym);
	}
	void Reset() {m_mainPool.Reset(); m_reservePool.Reset();}
};

// Will compute the value of IsInside once, and reuse the answer for subsequent calls
class BoundingBoxProperties
{
	Vector<4> m_avgModelPos;
	float m_radSquared;
	bool m_isInside;

	public:
	BoundingBoxProperties(SymbolInstance* curInstance, const Matrix<4,4>& modelView);

	const Vector<4>& GetAvgModelPos() const {return m_avgModelPos;}
	float GetRadSquared() const {return m_radSquared;}
	bool IsInside() const {return m_isInside;}
};

// Takes an analyzer list and transforms it to the corresponding CustomParam list
template <class T>
struct CustomParamListTransform
{
};

template <class H, class R>
struct CustomParamListTransform< TemplateList<H*, R> >
{
	typedef TemplateList<typename H::CustomParams, typename CustomParamListTransform<R>::ToParams> ToParams;
};

template <>
struct CustomParamListTransform< ListNull >
{
	typedef ListNull ToParams;
};

class WorldInstance
{
	struct ExtraInstanceInfo
	{
		ExtraInstanceInfo()
		{};

		ExtraInstanceInfo(int pLayer, Matrix<4,4>& pModelView)
			: layer(pLayer), modelView(pModelView)
		{};

		int layer; // Which layer it belongs to, layer 0 is the closest to us.  Layers are created to get by floating point errors on large numbers
		Matrix<4,4,false> modelView; // Pre-calculated modelview matrix for this instance
	};

	typedef std::map<SymbolInstance*, ExtraInstanceInfo > ExtraInfoMap;
	struct StaticTraversalInfo // Info that doesn't change while we are Instancing down
	{
		SymbolInstance* origInst;

		ExtraInfoMap extraInstanceInfo;
	};
	StaticTraversalInfo m_travInfo;
	SymbolInstance* m_nextRefInstance;


	const WorldDefinition& m_worldDef;

	static const int NUM_POOLS=2;
	SymbolInstancePool* m_instancePools[NUM_POOLS];
	int m_curPool;
	void UseNextPool();
	SymbolInstancePool& GetCurPool() {return *m_instancePools[m_curPool];}

	SymbolInstance* m_refInstance;

	Matrix<4,4> FindRelationMatrix(SymbolInstance* source, SymbolInstance* dest);
	int ChooseChildIndexForParent(Production* prod, Symbol* sym, const Vector<4>& curPos);

	static const int MAX_UP_INSTANCES = 18;

	public:
	WorldInstance(const WorldDefinition& worldDef, int poolSize, Symbol* startSym = 0);
	~WorldInstance();

	template <class ANALLIST> // haha
	void InstanceDownWithinPool(SymbolInstance* parent, const Matrix<4,4>& modelView, ANALLIST& analyzerList, const typename CustomParamListTransform<ANALLIST>::ToParams& customParams, int layer, int recurseCount);
	template <class ANALLIST>
	void InstanceSymbols(const Matrix<4,4>& modelView, ANALLIST& analyzerList);
	
	Matrix<4,4> SetNewRefInstance(SymbolInstance* nextInstance);

	Production* ChooseProductionForNewSymbolInstance(Symbol* sym);
};



template <class ANALLIST>
void WorldInstance::InstanceSymbols(const Matrix<4,4>& modelView, ANALLIST& analyzerList)
{
	// Flip our instance storage
	UseNextPool();

	Matrix<4,4> curModelView(modelView);
	
	Vector<4> curPos(curModelView.GetColumn(3));

	std::list<SymbolInstance*> tempSyms;
	SymbolInstance* curInstance = m_refInstance;
	SymbolInstance* srcRefInstance = m_refInstance; // If we need to copy m_refInstance, this will point to the copy, and if not, it will point to m_refInstance

	m_travInfo.extraInstanceInfo.clear();

	int curLayer = 0;

	// Constructed by multiplying the eigenvalues of the matrices we traverse together
	// used to determine when to stop moving up the tree
	double inverseContractionFactor=1.0f;

	curInstance->m_mustReinstance = true;
	m_travInfo.extraInstanceInfo[curInstance] = ExtraInstanceInfo(curLayer, curModelView);

	// Start heading to the tree's root (although it may not really have one if it's infinite, in which case we go up MAX_UP_INSTANCES times)
	for(int i=0; i<MAX_UP_INSTANCES; ++i)
	{
		const Production::ChildDescriptor* meAsAChild = 0;

		if(curInstance->m_parent != 0)
		{
			meAsAChild = curInstance->GetMyChildDescriptorFromParent();
	
			ASSERT(meAsAChild != 0);

			if(curInstance == m_refInstance)
			{
				srcRefInstance = curInstance;
			}

			curInstance = curInstance->m_parent;
		}
		else
		{
			if(curInstance->m_sym->GetRSProductions().size() == 0) break; // No possible parent... done.

			// We'll have to create our parent then, things get interesting here
			SymbolInstance* newParent = new SymbolInstance[1];
			tempSyms.push_back(newParent);

			// First we randomly choose a production to use
			Production* randomProd = curInstance->m_sym->GetRSProductionsArray()[rand() % curInstance->m_sym->GetRSProductions().size()];

			// Now we figure out which child of that production we are by choosing
			// the one that puts our position closest to the center of the parent
			int childIndex = ChooseChildIndexForParent(randomProd, curInstance->m_sym, curPos);

			
			newParent->m_prod = randomProd;
			newParent->m_sym = randomProd->GetLS();
			newParent->m_parent = 0;
			newParent->m_hasChildren = true;
			newParent->m_numChildren = randomProd->GetChildren().size();
			
			// Now make the children
			newParent->m_children = new SymbolInstance[newParent->m_numChildren];
			tempSyms.push_back(newParent->m_children);
			
			bool foundChild = false;
			int k=0;
			for(Production::ChildList::const_iterator iter = randomProd->GetChildren().begin();
				iter != randomProd->GetChildren().end(); ++iter, ++k)
			{
				if(k == childIndex) // If this is the index of the child we're coming from, just hook the child up instead of creating a new one
				{
					ASSERT(meAsAChild == 0);
					meAsAChild = *iter;
					newParent->m_children[k] = *curInstance;
					if(newParent->m_children[k].m_hasChildren)
					{
						for(int j=0; j<newParent->m_children[k].m_numChildren; ++j)
						{
							newParent->m_children[k].m_children[j].m_parent = &newParent->m_children[k];
						}
					}

					ASSERT(newParent->m_children[k].m_parent == 0);
					newParent->m_children[k].m_parent = newParent;
					
					foundChild = true;
						
					if(curInstance == m_refInstance)
					{
						srcRefInstance = &newParent->m_children[k];
					}

					// Re-add this guy in to the precalculated stuff
					ExtraInfoMap::const_iterator found = m_travInfo.extraInstanceInfo.find(curInstance);
					ASSERT(found != m_travInfo.extraInstanceInfo.end());
					m_travInfo.extraInstanceInfo[&newParent->m_children[k]] = found->second;
				}
				else
				{				
					// Generate the children
					newParent->m_children[k].m_sym = (*iter)->sym;
					newParent->m_children[k].m_hasChildren = false;
					newParent->m_children[k].m_parent = newParent;
					newParent->m_children[k].m_prod = ChooseProductionForNewSymbolInstance((*iter)->sym);
				}
			}
			ASSERT(foundChild);
			
			curInstance = newParent;
		}

		ASSERT(meAsAChild != 0);
		
		curPos = meAsAChild->transform * curPos;
		
		curModelView = curModelView*meAsAChild->invTransform;

		// Figure out how this affects our contraction factor
		double minMagEigenvalue = abs(meAsAChild->eigenValues[0]);
		for(int i=1; i<3; ++i)
		{
			double curMagEigenvalue = abs(meAsAChild->eigenValues[i]);
			if(curMagEigenvalue < minMagEigenvalue) minMagEigenvalue = curMagEigenvalue;
		}
		inverseContractionFactor /= minMagEigenvalue; 

		const double MAX_INVERSE_CONTRACTION_FACTOR = 300; // We will draw things up to this many times bigger(along 1 axis) than our current object
		if(inverseContractionFactor > MAX_INVERSE_CONTRACTION_FACTOR)
		{
			// We're getting too big, reset the modelview's scale, and start drawing in another layer
			curModelView *= (float)(1/inverseContractionFactor); // Kill the scale in the matrix
			curModelView.GetVal(3,3) = 1.0f;			

			inverseContractionFactor = 1.0f;

			++curLayer;
		}

		curInstance->m_mustReinstance = true;
		m_travInfo.extraInstanceInfo[curInstance] = ExtraInstanceInfo(curLayer, curModelView);
	}

	curInstance->m_parent = 0;
	
	ASSERT(srcRefInstance != 0);
	SymbolInstance* curSym = srcRefInstance;
	while(curSym != 0)
	{
		ASSERT(curSym->m_mustReinstance);
		if(curSym->m_parent == 0) break;
		
		curSym = curSym->m_parent;
	}
	ASSERT(curSym == curInstance);
	
	SymbolInstance* root = GetCurPool().MakeInstance();
	*root = *curInstance;
	if(root->m_hasChildren)
	{
		for(int i=0; i<root->m_numChildren; ++i)
		{
			root->m_children[i].m_parent = root;
		}
	}
		
	// Setup a structure containing parameters for InstanceDownWithinPool that we won't change
	m_travInfo.origInst = srcRefInstance;

	ASSERT(curInstance->m_mustReinstance);
	ASSERT(srcRefInstance->m_mustReinstance);

	m_nextRefInstance = (curInstance == srcRefInstance ? root : 0);

	ExtraInfoMap::const_iterator found = m_travInfo.extraInstanceInfo.find(curInstance);
	ASSERT(found != m_travInfo.extraInstanceInfo.end());

	static const int MAX_RECURSE_COUNT = MAX_UP_INSTANCES + 30;
	CustomParamListTransform<ANALLIST>::ToParams initialParams;
	InstanceDownWithinPool(root, Matrix<4,4>(found->second.modelView), analyzerList, initialParams, found->second.layer, MAX_RECURSE_COUNT);

	ASSERT(m_nextRefInstance != 0);
	ASSERT(GetCurPool().Contains(m_nextRefInstance));

	m_refInstance = m_nextRefInstance;
	
	for(std::list<SymbolInstance*>::iterator iter = tempSyms.begin(); iter != tempSyms.end(); ++iter)
	{
		delete[] *iter;
	}
}

template <class ANALLIST>
static bool AnalyzeAnalyzerList(ANALLIST& analyzerList, const typename CustomParamListTransform<ANALLIST>::ToParams& in, typename CustomParamListTransform<ANALLIST>::ToParams& out, SymbolInstance* parent, const Matrix<4,4>& modelView, const BoundingBoxProperties& curProps, int layer)
{
	return analyzerList.head->Analyze(parent, modelView, curProps, in.head, out.head, layer)
		| AnalyzeAnalyzerList(analyzerList.rest, in.rest, out.rest, parent, modelView, curProps, layer);
}

template <>
static bool AnalyzeAnalyzerList<ListNull>(ListNull& analyzerList, const ListNull& in, ListNull& out, SymbolInstance* parent, const Matrix<4,4>& modelView, const BoundingBoxProperties& curProps, int layer)
{
	return false;
}

template <class ANALLIST>
void WorldInstance::InstanceDownWithinPool(SymbolInstance* parent, const Matrix<4,4>& modelView, ANALLIST& analyzerList, const typename CustomParamListTransform<ANALLIST>::ToParams& customParams, int layer, int recurseCount)
{
	ASSERT(GetCurPool().Contains(parent));

	// Iterate through the analyzers and let them operate on the current instance
	BoundingBoxProperties curProps(parent, modelView);

	CustomParamListTransform<ANALLIST>::ToParams resultingParams;
	bool recurseMore = AnalyzeAnalyzerList(analyzerList, customParams, resultingParams, parent, modelView, curProps, layer);

	bool recursed = false;
	// If the area is too small, render the approximation and stop instancing
	if(recurseCount > 0 && (recurseMore || (parent->m_mustReinstance && parent->m_prod !=0)) )
	{
		ASSERTMSG(parent->m_prod != 0, "All symbols with 0 productions should have resulted in recurseMore being false");

		if(!parent->m_hasChildren)
		{
			// Parent has no children, so we gotta make some out of our production
			parent->m_numChildren = parent->m_prod->GetChildren().size();
		}

		// Instance the children
		SymbolInstance* newChildren = GetCurPool().MakeInstanceArray(parent->m_numChildren, parent->m_mustReinstance);
		if(parent->m_mustReinstance) ASSERT(newChildren);
		
		if(newChildren)
		{
			recursed = true;
			
			for(unsigned i=0; i<parent->m_prod->GetChildren().size(); ++i)
			{
				SymbolInstance& curInstance = newChildren[i];

				
				if(parent->m_hasChildren)
				{
					curInstance = parent->m_children[i];
					curInstance.m_parent = parent;
					if(&parent->m_children[i] == m_travInfo.origInst)
					{
						ASSERT(m_nextRefInstance == 0);
						m_nextRefInstance = &curInstance;
					}
				}
				else
				{
					curInstance.m_sym = parent->m_prod->GetChildren()[i]->sym;
					curInstance.m_parent = parent;
					curInstance.m_hasChildren = false;
					curInstance.m_prod  = ChooseProductionForNewSymbolInstance(curInstance.m_sym);
				}
				
				// See if the modelview matrix was previously calculated while walking up the tree
				ExtraInfoMap::const_iterator found = m_travInfo.extraInstanceInfo.find(&parent->m_children[i]);

				InstanceDownWithinPool(&curInstance, (found == m_travInfo.extraInstanceInfo.end()
													? modelView*parent->m_prod->GetChildren()[i]->transform
													: Matrix<4,4>(found->second.modelView)),
													analyzerList, resultingParams,
													(found == m_travInfo.extraInstanceInfo.end()
													? layer : found->second.layer), recurseCount-1);
			}

			parent->m_children = newChildren;
			parent->m_hasChildren = true;
		}
	}
	if(!recursed)
	{
		parent->m_hasChildren = false;

	}
	
	parent->m_mustReinstance = false;
}


#endif
