#include "SceneExplorer.h"

#include "Camera.h"
#include "WorldDefinition.h"
#include "WorldInstance.h"
#include "RenderAnalyzer.h"

#include <System.h>
#include <R3Utils.h>

#include <GraphicsFrame.h>
#include <Projection.h>

#include <math.h>
#include <iostream>
#include <boost/bind.hpp>

#ifdef PLAT_UNIX
#define USE_SDL
#endif

#ifdef USE_SDL
#include <SDLWindow.h>
#endif

SceneExplorer::SceneExplorer(const WorldDefinition& worldDef, GraphicsContext& graphContext, KeyboardInput& keyboardInput, MouseInput& mouseInput)
: m_forward(false)
, m_back(false)
, m_left(false)
, m_right(false)
, m_quitFlag(false)
, m_mouseX(0)
, m_mouseY(0)
, m_worldDef(worldDef)
, m_graphicsContext(graphContext)
, m_keyboardInput(keyboardInput)
, m_mouseInput(mouseInput)
{
	if(worldDef.GetSymbols().size() == 0)
	{
		throw GeneralException("WorldDefinition must have at least 1 symbol in it.");
	}
	
	m_sceneThread = new Thread(boost::bind(&SceneExplorer::Run, this));
}

SceneExplorer::~SceneExplorer()
{
	m_quitFlag = true;
	delete m_sceneThread;
}

void SceneExplorer::ProcessKeyboardInput(const KeyboardInput::KeyEvent& key)
{
	switch(key.key)
	{
		case KeyboardInput::KT_W:
		{
			m_forward = key.pressed;
		} break;
		case KeyboardInput::KT_S:
		{
			m_back = key.pressed;
		} break;
		case KeyboardInput::KT_A:
		{
			m_left = key.pressed;
		} break;
		case KeyboardInput::KT_D:
		{
			m_right = key.pressed;
		} break;
		
		default: break;
	};
}

void SceneExplorer::ProcessMouseMove(const MouseInput::MoveEvent& moveEvent)
{
	m_mouseX += moveEvent.dx;
	m_mouseY += moveEvent.dy;
}

void SceneExplorer::ProcessInput(float deltaT, Camera& cam, float curScale)
{
	const float speed = 0.4f * curScale;

	if(m_forward)
	{
		Vector<4> forward(cam.GetRotMatrix().GetColumn(2));
		cam.SetPosition(cam.GetPosition() + forward * deltaT * speed);
	}
	if(m_back)
	{
		Vector<4> back(-cam.GetRotMatrix().GetColumn(2));
		cam.SetPosition(cam.GetPosition() + back * deltaT * speed);
	}
	if(m_left)
	{
		Vector<4> left(cam.GetRotMatrix().GetColumn(0));
		cam.SetPosition(cam.GetPosition() + left * deltaT * speed / 2);
	}
	if(m_right)
	{
		Vector<4> right(-cam.GetRotMatrix().GetColumn(0));
		cam.SetPosition(cam.GetPosition() + right * deltaT * speed / 2);
	}


	cam.AddPitch((float)m_mouseY / 1000.0f);
	cam.AddYaw((float)m_mouseX / 1000.0f);

	m_mouseY = 0;
	m_mouseX = 0;
}

float NormalizeScale(float deltaT, float curScale)
{
	if(curScale > 1.0f) return 1.0f; // Never go faster than we should

	const float DECREASE_PER_SECOND = 0.4f; // Reduce difference by a half every second

	float diff = curScale - 1.0f;

	float newDiff = diff * pow(1-DECREASE_PER_SECOND, deltaT);

	return 1.0f + newDiff;
}


class IdealRefAnalyzer
{
	float m_bestScore;
	SymbolInstance* m_bestInstance;

	Mutex m_monitorLock;

	public:
	IdealRefAnalyzer()
		: m_bestInstance(0)
	{
	};
	~IdealRefAnalyzer()
	{
	}

	SymbolInstance* GetBestInstance() {return m_bestInstance;}

	struct CustomParams
	{
		CustomParams()
		{};
	};

	bool Analyze(SymbolInstance* curInstance, const Matrix<4,4>& modelView, const BoundingBoxProperties& boxProps, const CustomParams& inParams, CustomParams& outParams, int layer)
	{
		if(layer != 0) return false; // Only let layer 0 items be candidates for a reference object, since layers above 0 are definitely bigger than what we want

		float score;
		if(boxProps.IsInside())
		{
			score = 1/boxProps.GetRadSquared();
			
			const float MIN_SCORE = 0.000000001f;
			if(score < MIN_SCORE) score = MIN_SCORE;
		}
		else
		{
			score = -100000 + boxProps.GetRadSquared();
			if(score > -1) score = -1;
		}

		Mutex::ScopedLock lock(m_monitorLock);

		if(m_bestInstance == 0 || score > m_bestScore)
		{
			m_bestScore = score;
			m_bestInstance = curInstance;
		}

		return false;
	}
};

void SceneExplorer::Run()
{
	try
	{
		KeyboardInput::KeyListenerConnection keyListener(m_keyboardInput, boost::bind<void>(&SceneExplorer::ProcessKeyboardInput, this, _1));
		MouseInput::MoveListenerConnection mouseListener(m_mouseInput, boost::bind<void>(&SceneExplorer::ProcessMouseMove, this, _1));

		WorldInstance worldInst(m_worldDef, 2000000, 0);

		Vector<4, VT_INSTANCE, false> lightPosVec(1, 3, 0.25, 0);
		
		int framesSincePrint = 0;
		float timeSincePrint = 0.0f;

		const int NUM_FRAMES = 2;
		std::auto_ptr<GraphicsFrame> frames[NUM_FRAMES]; // One back frame and one front frame
		int curFrame = 0;

		float curScale = 1.0f; // Current scale for the movement speed, so we can ease the user in to other frames of reference

		Camera myCam(MakeTranslationMatrix(Vector<3>(M_PI, M_PI, M_PI)));
		long long curTick = System::GetTicks();
		while(!m_quitFlag)
		{
			long long nextTick = System::GetTicks();
			float deltaT = (float)(nextTick - curTick) / System::GetTicksPerSecond();

			curTick = nextTick;

			++framesSincePrint;
			timeSincePrint += deltaT;

			if(timeSincePrint > 1.0f)
			{
				std::cout << "FPS: " << framesSincePrint / timeSincePrint << std::endl;
				framesSincePrint = 0;
				timeSincePrint = 0;
			}

			ProcessInput(deltaT, myCam, curScale);

			ASSERT(frames[curFrame].get() == 0);
			frames[curFrame] = std::auto_ptr<GraphicsFrame>(new GraphicsFrame(100000, 120000));

			GraphicsFrame& workFrame = *frames[curFrame];

			workFrame.SetLightDir(myCam.GetModelView()*lightPosVec);

			const float nearP = 0.001f;
			const float farP = 600.0f;
			Matrix<4,4> projMatrix = MakePerspectiveProjectionMatrix(45.0f, (float)m_graphicsContext.GetWidth()/m_graphicsContext.GetHeight(), nearP, farP);
			IdealRefAnalyzer idealRefAnalyzer;
			RenderAnalyzer renderAnalyzer(workFrame, projMatrix, nearP);

			worldInst.InstanceSymbols(myCam.GetModelView(), MakeTemplateList(&idealRefAnalyzer, &renderAnalyzer));

			// Start rendering the new frame
			workFrame.Lock();
			m_graphicsContext.RenderGraphicsFrame(workFrame, GraphicsContext::RenderOptions(projMatrix));

			// Okay, now make a new camera relative to the instance suggested by the instancing system
			// And start drawing off of that one
			Matrix<4,4> relMatrix = worldInst.SetNewRefInstance(idealRefAnalyzer.GetBestInstance());
			
			lightPosVec = relMatrix*lightPosVec;
			lightPosVec.Normalize();
			
			myCam = Camera(relMatrix * myCam.GetTransMatrix());
			Tuple<std::complex<double>, 3> ret = GetEigenvaluesOfLinear(relMatrix);

			curScale *= (float)(abs(ret[0]));
			curScale = NormalizeScale(deltaT, curScale); // Bring the curScale to 1 over time

			// Wrap up the old frame
			++curFrame;
			curFrame = curFrame % NUM_FRAMES;
			frames[curFrame].reset();
		}
	}
	catch(std::exception& e)
	{
		std::cerr << "Error: " << e.what() << std::endl;
	}
}


