#include "WorldDefinition.h"
#include <ScopedFStream.h>

#include <Model.h>

#include <map>

#include <boost/bind.hpp>

void Symbol::AddLSProduction(Production* prod)
{
	ASSERT(m_lsProductions.find(prod) == m_lsProductions.end());
	m_lsProductions.insert(prod);
	m_lsProductionsArray.push_back(prod);
}

void Symbol::RemoveLSProduction(Production* prod)
{
	ASSERT(m_lsProductions.find(prod) != m_lsProductions.end());
	m_lsProductions.erase(prod);

	bool found = false;
	ProductionArray::iterator iter;
	for(iter = m_lsProductionsArray.begin(); 
		iter != m_lsProductionsArray.end(); ++iter)
	{
		if(*iter == prod)
		{
			m_lsProductionsArray.erase(iter);
			found = true;
			break;
		}
	}
	ASSERT(found);
}

void Symbol::AddRSProduction(Production* prod)
{
	if(m_rsProductions.find(prod) == m_rsProductions.end())
	{
		m_rsProductions.insert(prod);
		m_rsProductionsArray.push_back(prod);
	}
}
void Symbol::RemoveRSProduction(Production* prod)
{
	ASSERT(m_rsProductions.find(prod) != m_rsProductions.end());
	m_rsProductions.erase(prod);

	bool found = false;
	ProductionArray::iterator iter;
	for(iter = m_rsProductionsArray.begin(); 
		iter != m_rsProductionsArray.end(); ++iter)
	{
		if(*iter == prod)
		{
			m_rsProductionsArray.erase(iter);
			found = true;
			break;
		}
	}
	ASSERT(found);
}

static Production::ChildDescriptor MakeDescriptor(const Matrix<4,4>& transform, Symbol* sym)
{
	Production::ChildDescriptor ret;

	ret.sym = sym;

	ret.eigenValues = GetEigenvaluesOfLinear(transform);
	for(int i=1; i<3; ++i)
	{
		// Make sure that this is a uniform scale (along with any kind of rotation)
		ASSERT(NumbersClose(abs(ret.eigenValues[0]), abs(ret.eigenValues[i]), 0.1)); // Make epsilon stricter when we have better tools for creating productions
	}

	ret.transform = transform;
	ret.invTransform = transform.Inverse();

	return ret;
}

void Production::AddRS(const Matrix<4,4>& transform, Symbol* sym)
{
	m_children.push_back(new ChildDescriptor(MakeDescriptor(transform, sym)));

	// Add this production to the enumeration of productions the given symbol is in the right side of
	if(sym->GetRSProductions().find(this) == sym->GetRSProductions().end())
	{
		sym->AddRSProduction(this);
	}
}

void Production::EditRS(int childIndex, const Matrix<4,4>& transform, Symbol* sym)
{
	*(m_children[childIndex]) = MakeDescriptor(transform, sym);
}

void Production::RemoveRS(int childIndex)
{
	ChildList::iterator iter;
	int i=0;
	for(iter = m_children.begin(); iter != m_children.end(); ++iter, ++i)
	{
		if(i == childIndex) break;
	}
	ASSERT(iter != m_children.end());

	Symbol* removedSym = (*iter)->sym;
	m_children.erase(iter);

	// And now erase it from the appropriate symbol's right hand side if there are no other symbols like it on the RS
	bool stillThere = false;
	for(iter = m_children.begin(); iter != m_children.end(); ++iter)
	{
		if((*iter)->sym == removedSym)
		{
			stillThere = true;
			break;
		}
	}
	if(!stillThere)
	{
		removedSym->RemoveRSProduction(this);
	}
}

WorldDefinition::WorldDefinition(ResourceManager& resMan)
: m_resMan(resMan)
{
}

WorldDefinition::~WorldDefinition()
{
	for(ProductionSet::iterator i = m_productions.begin(); i != m_productions.end(); ++i) delete *i;
	for(SymbolSet::iterator i = m_symbols.begin(); i != m_symbols.end(); ++i) delete *i;
}

void WorldDefinition::AddSymbol(std::auto_ptr<Symbol> sym)
{
	Symbol* mySym = sym.release();

	ASSERT(m_symbols.find(mySym) == m_symbols.end());

	m_symbols.insert(mySym);

	m_symArray.push_back(mySym);
}

void WorldDefinition::RemoveSymbol(Symbol* sym)
{
	ASSERT(sym->GetLSProductions().size() == 0);
	ASSERT(sym->GetRSProductions().size() == 0);

	ASSERT(m_symbols.find(sym) != m_symbols.end());
	m_symbols.erase(sym);

	bool found = false;
	for(SymbolArray::iterator iter = m_symArray.begin(); iter != m_symArray.end(); ++iter)
	{
		if(*iter == sym)
		{
			found = true;
			m_symArray.erase(iter);
			break;
		}
	}

	delete sym;
}

void WorldDefinition::AddProduction(std::auto_ptr<Production> prod)
{
	Production* myProd = prod.release();

	ASSERT(m_symbols.find(myProd->GetLS()) != m_symbols.end());

	m_productions.insert(myProd);
	m_prodArray.push_back(myProd);

	// Tell the symbol that it is now on the left side of this production
	ASSERT(myProd->GetLS()->m_lsProductions.find(myProd) == myProd->GetLS()->m_lsProductions.end());
	myProd->GetLS()->AddLSProduction(myProd);

	// Add this production to the symbol classes that it involves
	for(Production::ChildList::iterator iter = myProd->m_children.begin(); iter != myProd->m_children.end(); ++iter)
	{
		Symbol* curSym = (*iter)->sym;

		if(curSym->m_rsProductions.find(myProd) == curSym->m_rsProductions.end())
		{
			curSym->AddRSProduction(myProd);
		}
	}
}

void WorldDefinition::RemoveProduction(Production* prod)
{
	ProductionSet::iterator foundInProductions = m_productions.find(prod);
	ASSERT(foundInProductions != m_productions.end());
	m_productions.erase(foundInProductions);

	bool foundInArray = false;
	for(ProductionArray::iterator iter = m_prodArray.begin(); iter != m_prodArray.end(); ++iter)
	{
		if(prod == *iter)
		{
			foundInArray = true;
			m_prodArray.erase(iter);
			break;
		}
	}
	ASSERT(foundInArray);

	// Remove the links from the left side symbol it is associated with
	prod->GetLS()->RemoveLSProduction(prod);

	// Remove the links from all the symbols that are on the right hand side of this production
	std::set<Symbol*> rsSymSet; // To track what we've removed already so we don't remove the production twice from one symbol
	Production::ChildList& children = prod->GetChildren();
	for(Production::ChildList::iterator iter = children.begin(); iter != children.end(); ++iter)
	{
		Symbol* curSym = (*iter)->sym;

		if(rsSymSet.find(curSym) != rsSymSet.end()) continue; // We've already erased this one
		rsSymSet.insert(curSym);

		// Remove!
		curSym->RemoveRSProduction(prod);
	}

	delete prod;
}

Symbol* WorldDefinition::GetRandomSymbol() const
{
	ASSERT(m_symArray.size() > 0);

	return m_symArray[rand() % m_symArray.size()];
}

Production* WorldDefinition::GetRandomProduction() const
{
	ASSERT(m_prodArray.size() > 0);

	return m_prodArray[rand() % m_prodArray.size()];
}

bool WorldDefinition::Contains(Symbol* sym) const
{
	return m_symbols.find(sym) != m_symbols.end();
}

bool WorldDefinition::Contains(Production* prod) const
{
	return m_productions.find(prod) != m_productions.end();
}




class SymbolsFile
{
public:
	struct Symbol
	{
		std::string name;
		float threshold;
		std::string modelFile;
	};

	struct Production
	{
		Symbol* ls;
		std::string rsFile;
		::Production::FadeMode fadeMode;
	};
private:

	template<class Container>
	struct ItemList
	{
		Container items;

		~ItemList()
		{
			for(Container::iterator iter = items.begin(); iter != items.end(); ++iter)
			{
				delete *iter;
			}
		}

		void ClearRelease()
		{
			items.clear();
		}

		void TakeFrom(ItemList<Container>& other)
		{
			items = other.items;
			other.items.clear();
		}
	};

	ItemList<std::vector<Symbol*> > m_symbols;
	ItemList<std::vector<Production*> > m_prods;
public:
	class ParseException
	{
		int m_lineNum;
		std::string m_msg;
	public:
		ParseException(int lineNum, const std::string& msg)
			: m_lineNum(lineNum)
			, m_msg(msg)
		{};

		int GetLineNum() const {return m_lineNum;}
		const std::string& GetMessage() const {return m_msg;}
	};

	typedef std::vector<Symbol*> SymbolList;
	typedef std::vector<Production*> ProdList;
	const SymbolList& GetSymbols() const {return m_symbols.items;}
	const ProdList& GetProductions() const {return m_prods.items;}

	Symbol* FindSymbol(const std::string& symName) const
	{
		for(SymbolList::const_iterator iter = m_symbols.items.begin(); iter != m_symbols.items.end(); ++iter)
		{
			if((*iter)->name == symName)
			{
				return *iter;
			}
		}

		return 0;
	}

	SymbolsFile(std::ifstream& in)
	{
		ItemList<SymbolList> symList;
		ItemList<ProdList> prodList;

		int lineNumber = 0;

		while(!in.eof())
		{
			++lineNumber;

			char lineBuf[500];
			in.getline(lineBuf, 500);

			if(lineBuf[0] == '#') continue; // ignore comments

			std::istringstream liness(lineBuf);

			std::string curTok;
			liness >> curTok;
			if(liness.eof()) continue;

			std::string symName = curTok;

			liness >> curTok;
			if(curTok == "=")
			{
				// Parse this symbol definition
				Symbol sym;

				sym.name = symName;

				while(!liness.eof())
				{
					liness >> curTok;
					if(curTok.size() < 2 || curTok[1] != ':') throw ParseException(lineNumber, "Expected ':' sign");

					if(curTok[0] == 'T')
					{
						std::string thresholdStr = curTok.substr(2);
						std::istringstream threshss(thresholdStr);

						threshss >> sym.threshold;
					}
					else if(curTok[0] == 'M')
					{
						sym.modelFile = curTok.substr(2);
					}
					else
					{
						throw ParseException(lineNumber, "Unexpected symbol definition command");
					}
				}

				symList.items.push_back(new Symbol(sym));
			}
			else if(curTok == "->")
			{
				// Parse this production definition
				
				Symbol* foundSym = 0;
				// Find the symbol this refers to
				for(SymbolList::const_iterator iter=symList.items.begin(); iter != symList.items.end(); ++iter)
				{
					if((*iter)->name == symName)
					{
						foundSym = *iter;
						break;
					}
				}
				if(foundSym == 0) throw ParseException(lineNumber, "Could not find symbol");

				Production prod;
				prod.fadeMode = ::Production::FM_NONE;

				liness >> curTok;
				if(!liness.eof()) // Is this not an empty production?
				{
					if(curTok == "fadein")
					{
						prod.fadeMode = ::Production::FM_FADEIN;
					}
					else if(curTok == "fadeout")
					{
						prod.fadeMode = ::Production::FM_FADEOUT;
					}
					else if(curTok == "nofade")
					{
						prod.fadeMode = ::Production::FM_NONE;
					}
					else
					{
						throw ParseException(lineNumber, "Expected fadein or fadeout");
					}

					liness >> curTok;
					prod.rsFile = curTok;
				}

				prod.ls = foundSym;

				prodList.items.push_back(new Production(prod));
			}
			else
			{
				throw ParseException(lineNumber, "Expected '=' or '->' sign");
			}		
		}

		m_symbols.TakeFrom(symList);
		m_prods.TakeFrom(prodList);
	};
};
									
static void RenderModel(const Model* model, GraphicsFrame* frame, const Matrix<4,4>* modelView, float alpha, int layer)
{
	model->AddToGraphicsFrame(*frame, *modelView, alpha, layer);
}

void WorldDefinition::LoadFromDirectory(const std::string& dirName)
{
	SymbolsFile symbols(ScopedFStream<std::ifstream>(dirName + "/symbols.txt").s);

	std::map<const SymbolsFile::Symbol*, Symbol*> symMap; // Maps the symbols from the data file to their WorldDefinition version that they were used to create

	// Add our models to the model list
	for(SymbolsFile::SymbolList::const_iterator iter = symbols.GetSymbols().begin();
		iter != symbols.GetSymbols().end();
		++iter)
	{
		const SymbolsFile::Symbol& symFileSym = **iter;

		// Create the WorldDefinition::Symbol from the SymbolsFile::Symbol and add it
		std::auto_ptr<Symbol> addSym(new Symbol());
		addSym->SetAreaApproximateThreshold(symFileSym.threshold);

		if(symFileSym.modelFile.size() > 0)
		{
			m_modelRefs.push_back(ResourceReference<Model, ModelFromObjFactory>(m_resMan, ResourceIdentifier(dirName + "/" + symFileSym.modelFile)));
			addSym->SetApproximationRenderFunc(boost::bind(&RenderModel, m_modelRefs.back().GetPtr(), _1, _2, _3, _4));
		}

		symMap[&symFileSym] = addSym.get();
		m_loadedSymbols[symFileSym.name] = addSym.get();
		AddSymbol(addSym);
	}

	// Now start reading in the productions
	for(SymbolsFile::ProdList::const_iterator iter = symbols.GetProductions().begin();
		iter != symbols.GetProductions().end();
		++iter)
	{
		const SymbolsFile::Production& symFileProd = **iter;

		// Find the left side symbol
		std::auto_ptr<Production> prod(new Production(symMap[symFileProd.ls], symFileProd.fadeMode));

		if(symFileProd.rsFile.size() > 0)
		{
			// Now we need to open up the obj file and start adding the productions defined within it
			ResourceReference<OBJFile> prodObj(m_resMan, ResourceIdentifier(dirName + "/" + symFileProd.rsFile));

			for(int i=0; i<prodObj->GetNumObjects(); ++i)
			{
				const OBJFile::Object& curObj = prodObj->GetObjectItem(i);

				// Get rid of the .#### at the end of the object name
				std::istringstream namess(curObj.name);
				char objName[500];
				namess.getline(objName, 500, '.');

				// Extract the right side symbol
				const SymbolsFile::Symbol* rsSym = symbols.FindSymbol(objName);
				if(rsSym == 0)
				{
					std::ostringstream oss;
					oss << "Error: Symbol '" << curObj.name << "' was not defined, referenced in " << dirName + "/" + symFileProd.rsFile;
					throw LoadException(oss.str());
				}

				// Now finally we need to extract the transformation from the cube in the object
				Vector<4> x((curObj.vertices[1] - curObj.vertices[2]).GetElements()); x[3] = 0;
				Vector<4> z((curObj.vertices[6] - curObj.vertices[2]).GetElements()); z[3] = 0;
				Vector<4> y((curObj.vertices[3] - curObj.vertices[2]).GetElements()); y[3] = 0;
				Vector<4> pos(curObj.vertices[2].GetElements()); pos[3] = 1;

				// Okay we have the four columns of our matrix, put it together
				Matrix<4,4> curTransform;
				curTransform.SetColumn(0, x);
				curTransform.SetColumn(1, y);
				curTransform.SetColumn(2, z);
				curTransform.SetColumn(3, pos);

				prod->AddRS(curTransform, symMap[rsSym]);
			}
		}

		AddProduction(prod);
	}

}

