#ifndef PRODUCTIONEDITSCENE_H_
#define PRODUCTIONEDITSCENE_H_

#include <Events.h>
#include <GraphicsFrame.h>
#include <GraphicsContext.h>
#include "ToolWorldDefinition.h"

/**
 *	The ProductionEditScene is responsible for creating the 3D scene that
 *	will be viewed in the 4 views by the QuadViewPanel.
 *  It will draw the bounding boxes (if enabled) and the actual scene itself
 *  as well if the preview mode is not set to None.  It listens to the events
 *  generated by the ToolWorldDefinition, and whenever it detects that the
 *  WorldDefinition has changed, it will re-create the scene.  When it recreates
 *  the scene, it will fire off an event of it's own, accessible through
 *  GetSceneChangedEventPort(), which can be listened to by a scene viewport
 *	to know when it needs to re-render itself.
 */
class ProductionEditScene
{
public:
	typedef Event<void ()> SceneChangedEvent;

	/** PreviewMode indicates how the child Symbols of a production are to be drawn */
	enum PreviewMode
	{
		PM_APPROXIMATIONS, /**< Draw only the approximation of every child Symbol */
		PM_WIREFRAME, /**< Same as PM_APPROXIMATIONS, except the approximations are drawn in wireframe */
		PM_NONE, /**< Don't draw any preview of the child objects */
	};

private:
	ToolWorldDefinition& m_world;
	const ToolProduction* m_prod;

	int m_selected;
	const Production::ChildDescriptor* GetSelectedDesc() const {return m_prod->GetProduction()->GetChildren()[m_selected];}

	PreviewMode m_previewMode;
	bool m_drawBounds;

	bool m_showMarker;
	Vector<4> m_markerPos;

	SceneChangedEvent m_sceneChangedEvent;

	static const int NUM_FRAMES = 2;
	std::auto_ptr<GraphicsFrame> m_frames[NUM_FRAMES];
	int m_curFrame;

	Event<void (const ToolProduction*, int)>::Connection* m_prodRSAddedConn;
	void HandleProductionRSAdded(const ToolProduction* prod, int index);
	
	Event<void (const ToolProduction*)>::Connection* m_prodChangedConn;
	void HandleProductionChanged(const ToolProduction* prod);

	struct ObjectFrameAssets
	{
		Material white;
		Material red;
		Material green;
		Material blue;
		Material selectCol;

		LineSegment x;
		LineSegment y;
		LineSegment z;
		
		LineSegment others[9]; // The other line segments (that aren't x,y,z)

		ObjectFrameAssets();
		void AddToFrame(GraphicsFrame& toFrame, int modelView, bool selected=false);
	};
	ObjectFrameAssets m_objectFrameAssets;

	void AddSceneHelpers(GraphicsFrame& toFrame); // Adds all the bounding frames and ground-grid

	void DoPreviewModeApproximations(GraphicsFrame& curFrame, bool wireframe);

	Event<void ()> m_selectedChangedEvent;

	void PrivateSetSelected(int selected);

public:
	ProductionEditScene(ToolWorldDefinition& world, const ToolProduction* prod);
	~ProductionEditScene();

	Event<void ()>::Port GetSceneChangedEventPort()
	{
		return m_sceneChangedEvent.GetPort();
	}

	Event<void ()>::Port GetSelectionChangedPort()
	{
		return m_selectedChangedEvent.GetPort();
	}

	void RenderScene(GraphicsContext& context, const GraphicsContext::RenderOptions& options);

	void PlaceMarker(const Vector<4>& pos);

	bool HaveSelection() const {return m_selected != -1;}
	const Production::ChildDescriptor* GetSelected() const {ASSERT(HaveSelection()); return GetSelectedDesc();}

	void MoveSelected(const Vector<4>& moveVec);
	void ScaleSelected(float scale);
	void RotateSelected(const UnitQuaternion& rot);

	void TrySelectItem(const Vector<4>& rayStart, const Vector<4>& otherPointOnRay);

	PreviewMode GetPreviewMode() const {return m_previewMode;}	/**< @return The currently set PreviewMode */
	bool GetDrawBounds() const {return m_drawBounds;} /**< @return The currently set option to draw child symbol bounding boxes or not */

	void SetPreviewMode(PreviewMode mode); /**< @param mode The new PreviewMode that should be used.  If different from GetPreviewMode(), will cause a scene redraw event to fire. */
	void SetDrawBounds(bool drawBounds); /**< @param drawBounds If set to true, will draw bounding boxes around all child Symbols.  If different from GetDrawBounds(), will cause a scene redraw event to fire. */

	/** Manually sets the selected component.  
	 * @param The component child index that you would like to select.
	 */
	void SetSelected(int selected);

	int GetSelectedIndex() const {return m_selected;} /**< @return The child index of the currently selected component.  -1 if none are selected. */

	/** Refreshes the scene by re-adding all of the draw calls. */
	void RecreateScene();
};

#endif
